"""
Send events from Docker events
:Depends:   Docker API >= 1.22
"""

import logging
import traceback

import salt.utils.event
import salt.utils.json

try:
    import docker  # pylint: disable=import-error,no-name-in-module
    import docker.utils  # pylint: disable=import-error,no-name-in-module

    HAS_DOCKER_PY = True
except ImportError:
    HAS_DOCKER_PY = False

log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# Default timeout as of docker-py 1.0.0
CLIENT_TIMEOUT = 60

# Define the module's virtual name
__virtualname__ = "docker_events"


def __virtual__():
    """
    Only load if docker libs are present
    """
    if not HAS_DOCKER_PY:
        return (False, "Docker_events engine could not be imported")
    return True


def start(
    docker_url="unix://var/run/docker.sock",
    timeout=CLIENT_TIMEOUT,
    tag="salt/engines/docker_events",
    filters=None,
):
    """
    Scan for Docker events and fire events

    Example Config

    .. code-block:: yaml

        engines:
          - docker_events:
              docker_url: unix://var/run/docker.sock
              filters:
                event:
                - start
                - stop
                - die
                - oom

    The config above sets up engines to listen
    for events from the Docker daemon and publish
    them to the Salt event bus.

    For filter reference, see https://docs.docker.com/engine/reference/commandline/events/
    """

    if __opts__.get("__role") == "master":
        fire_master = salt.utils.event.get_master_event(
            __opts__, __opts__["sock_dir"]
        ).fire_event
    else:
        fire_master = None

    def fire(tag, msg):
        """
        How to fire the event
        """
        if fire_master:
            fire_master(msg, tag)
        else:
            __salt__["event.send"](tag, msg)

    try:
        # docker-py 2.0 renamed this client attribute
        client = docker.APIClient(base_url=docker_url, timeout=timeout)
    except AttributeError:
        # pylint: disable=not-callable
        client = docker.Client(base_url=docker_url, timeout=timeout)
        # pylint: enable=not-callable

    try:
        events = client.events(filters=filters)
        for event in events:
            data = salt.utils.json.loads(
                event.decode(__salt_system_encoding__, errors="replace")
            )
            # https://github.com/docker/cli/blob/master/cli/command/system/events.go#L109
            # https://github.com/docker/engine-api/blob/master/types/events/events.go
            # Each output includes the event type, actor id, name and action.
            # status field can be ommited
            if data["Action"]:
                fire("{}/{}".format(tag, data["Action"]), data)
            else:
                fire("{}/{}".format(tag, data["status"]), data)
    except Exception:  # pylint: disable=broad-except
        traceback.print_exc()



import os.path
import posixpath

import salt.exceptions
import salt.loader
import salt.template
import salt.utils.args
import salt.utils.dictupdate
import salt.utils.stringio

CONTEXT_BASE = "slsutil"


def update(dest, upd, recursive_update=True, merge_lists=False):
    """
    Merge ``upd`` recursively into ``dest``

    If ``merge_lists=True``, will aggregate list object types instead of
    replacing. This behavior is only activated when ``recursive_update=True``.

    CLI Example:

    .. code-block:: shell

        salt '*' slsutil.update '{foo: Foo}' '{bar: Bar}'

    """
    return salt.utils.dictupdate.update(dest, upd, recursive_update, merge_lists)


def merge(obj_a, obj_b, strategy="smart", renderer="yaml", merge_lists=False):
    """
    Merge a data structure into another by choosing a merge strategy

    Strategies:

    * aggregate
    * list
    * overwrite
    * recurse
    * smart

    CLI Example:

    .. code-block:: shell

        salt '*' slsutil.merge '{foo: Foo}' '{bar: Bar}'
    """
    return salt.utils.dictupdate.merge(obj_a, obj_b, strategy, renderer, merge_lists)


def merge_all(lst, strategy="smart", renderer="yaml", merge_lists=False):
    """
    .. versionadded:: 2019.2.0

    Merge a list of objects into each other in order

    :type lst: Iterable
    :param lst: List of objects to be merged.

    :type strategy: String
    :param strategy: Merge strategy. See utils.dictupdate.

    :type renderer: String
    :param renderer:
        Renderer type. Used to determine strategy when strategy is 'smart'.

    :type merge_lists: Bool
    :param merge_lists: Defines whether to merge embedded object lists.

    CLI Example:

    .. code-block:: shell

        $ salt-call --output=txt slsutil.merge_all '[{foo: Foo}, {foo: Bar}]'
        local: {u'foo': u'Bar'}
    """

    ret = {}
    for obj in lst:
        ret = salt.utils.dictupdate.merge(ret, obj, strategy, renderer, merge_lists)

    return ret


def renderer(path=None, string=None, default_renderer="jinja|yaml", **kwargs):
    """
    Parse a string or file through Salt's renderer system

    .. versionchanged:: 2018.3.0
       Add support for Salt fileserver URIs.

    This is an open-ended function and can be used for a variety of tasks. It
    makes use of Salt's "renderer pipes" system to run a string or file through
    a pipe of any of the loaded renderer modules.

    :param path: The path to a file on Salt's fileserver (any URIs supported by
        :py:func:`cp.get_url <salt.modules.cp.get_url>`) or on the local file
        system.
    :param string: An inline string to be used as the file to send through the
        renderer system. Note, not all renderer modules can work with strings;
        the 'py' renderer requires a file, for example.
    :param default_renderer: The renderer pipe to send the file through; this
        is overridden by a "she-bang" at the top of the file.
    :param kwargs: Keyword args to pass to Salt's compile_template() function.

    Keep in mind the goal of each renderer when choosing a render-pipe; for
    example, the Jinja renderer processes a text file and produces a string,
    however the YAML renderer processes a text file and produces a data
    structure.

    One possible use is to allow writing "map files", as are commonly seen in
    Salt formulas, but without tying the renderer of the map file to the
    renderer used in the other sls files. In other words, a map file could use
    the Python renderer and still be included and used by an sls file that uses
    the default 'jinja|yaml' renderer.

    For example, the two following map files produce identical results but one
    is written using the normal 'jinja|yaml' and the other is using 'py':

    .. code-block:: jinja

        #!jinja|yaml
        {% set apache = salt.grains.filter_by({
            ...normal jinja map file here...
        }, merge=salt.pillar.get('apache:lookup')) %}
        {{ apache | yaml() }}

    .. code-block:: python

        #!py
        def run():
            apache = __salt__.grains.filter_by({
                ...normal map here but as a python dict...
            }, merge=__salt__.pillar.get('apache:lookup'))
            return apache

    Regardless of which of the above map files is used, it can be accessed from
    any other sls file by calling this function. The following is a usage
    example in Jinja:

    .. code-block:: jinja

        {% set apache = salt.slsutil.renderer('map.sls') %}

    CLI Example:

    .. code-block:: bash

        salt '*' slsutil.renderer salt://path/to/file
        salt '*' slsutil.renderer /path/to/file
        salt '*' slsutil.renderer /path/to/file.jinja default_renderer='jinja'
        salt '*' slsutil.renderer /path/to/file.sls default_renderer='jinja|yaml'
        salt '*' slsutil.renderer string='Inline template! {{ saltenv }}'
        salt '*' slsutil.renderer string='Hello, {{ name }}.' name='world'
    """
    if not path and not string:
        raise salt.exceptions.SaltInvocationError("Must pass either path or string")

    renderers = salt.loader.render(__opts__, __salt__)

    if path:
        path_or_string = __context__["fileclient"].get_url(
            path, "", saltenv=kwargs.get("saltenv", "base")
        )
    elif string:
        path_or_string = ":string:"
        kwargs["input_data"] = string

    ret = salt.template.compile_template(
        path_or_string,
        renderers,
        default_renderer,
        __opts__["renderer_blacklist"],
        __opts__["renderer_whitelist"],
        **kwargs,
    )
    return ret.read() if salt.utils.stringio.is_readable(ret) else ret


def _get_serialize_fn(serializer, fn_name):
    serializers = salt.loader.serializers(__opts__)
    fns = getattr(serializers, serializer, None)
    fn = getattr(fns, fn_name, None)

    if not fns:
        raise salt.exceptions.CommandExecutionError(
            f"Serializer '{serializer}' not found."
        )

    if not fn:
        raise salt.exceptions.CommandExecutionError(
            f"Serializer '{serializer}' does not implement {fn_name}."
        )

    return fn


def serialize(serializer, obj, **mod_kwargs):
    """
    Serialize a Python object using one of the available
    :ref:`all-salt.serializers`.

    CLI Example:

    .. code-block:: bash

        salt '*' --no-parse=obj slsutil.serialize 'json' obj="{'foo': 'Foo!'}

    Jinja Example:

    .. code-block:: jinja

        {% set json_string = salt.slsutil.serialize('json',
            {'foo': 'Foo!'}) %}
    """
    kwargs = salt.utils.args.clean_kwargs(**mod_kwargs)
    return _get_serialize_fn(serializer, "serialize")(obj, **kwargs)


def deserialize(serializer, stream_or_string, **mod_kwargs):
    """
    Deserialize a Python object using one of the available
    :ref:`all-salt.serializers`.

    CLI Example:

    .. code-block:: bash

        salt '*' slsutil.deserialize 'json' '{"foo": "Foo!"}'
        salt '*' --no-parse=stream_or_string slsutil.deserialize 'json' \\
            stream_or_string='{"foo": "Foo!"}'

    Jinja Example:

    .. code-block:: jinja

        {% set python_object = salt.slsutil.deserialize('json',
            '{"foo": "Foo!"}') %}
    """
    kwargs = salt.utils.args.clean_kwargs(**mod_kwargs)
    return _get_serialize_fn(serializer, "deserialize")(stream_or_string, **kwargs)


def boolstr(value, true="true", false="false"):
    """
    Convert a boolean value into a string. This function is
    intended to be used from within file templates to provide
    an easy way to take boolean values stored in Pillars or
    Grains, and write them out in the appropriate syntax for
    a particular file template.

    :param value: The boolean value to be converted
    :param true: The value to return if ``value`` is ``True``
    :param false: The value to return if ``value`` is ``False``

    In this example, a pillar named ``smtp:encrypted`` stores a boolean
    value, but the template that uses that value needs ``yes`` or ``no``
    to be written, based on the boolean value.

    *Note: this is written on two lines for clarity. The same result
    could be achieved in one line.*

    .. code-block:: jinja

        {% set encrypted = salt[pillar.get]('smtp:encrypted', false) %}
        use_tls: {{ salt['slsutil.boolstr'](encrypted, 'yes', 'no') }}

    Result (assuming the value is ``True``):

    .. code-block:: none

        use_tls: yes

    """

    if value:
        return true

    return false


def _set_context(keys, function, fun_args=None, fun_kwargs=None, force=False):
    """
    Convenience function to set a value in the ``__context__`` dictionary.

    .. versionadded:: 3004

    :param keys: The list of keys specifying the dictionary path to set. This
                 list can be of arbitrary length and the path will be created
                 in the dictionary if it does not exist.

    :param function: A python function to be called if the specified path does
                     not exist, if the force parameter is ``True``.

    :param fun_args: A list of positional arguments to the function.

    :param fun_kwargs: A dictionary of keyword arguments to the function.

    :param force: If ``True``, force the ```__context__`` path to be updated.
                  Otherwise, only create it if it does not exist.
    """

    target = __context__

    # Build each level of the dictionary as needed
    for key in keys[:-1]:
        if key not in target:
            target[key] = {}
        target = target[key]

    # Call the supplied function to populate the dictionary
    if force or keys[-1] not in target:
        if not fun_args:
            fun_args = []

        if not fun_kwargs:
            fun_kwargs = {}

        target[keys[-1]] = function(*fun_args, *fun_kwargs)


def file_exists(path, saltenv="base"):
    """
    Return ``True`` if a file exists in the state tree, ``False`` otherwise.

    .. versionadded:: 3004

    :param str path: The fully qualified path to a file in the state tree.
    :param str saltenv: The fileserver environment to search. Default: ``base``

    CLI Example:

    .. code-block:: bash

        salt '*' slsutil.file_exists nginx/defaults.yaml
    """

    _set_context(
        [CONTEXT_BASE, saltenv, "file_list"], __salt__["cp.list_master"], [saltenv]
    )
    return path in __context__[CONTEXT_BASE][saltenv]["file_list"]


def dir_exists(path, saltenv="base"):
    """
    Return ``True`` if a directory exists in the state tree, ``False`` otherwise.

    :param str path: The fully qualified path to a directory in the state tree.
    :param str saltenv: The fileserver environment to search. Default: ``base``

    .. versionadded:: 3004

    CLI Example:

    .. code-block:: bash

        salt '*' slsutil.dir_exists nginx/files
    """

    _set_context(
        [CONTEXT_BASE, saltenv, "dir_list"], __salt__["cp.list_master_dirs"], [saltenv]
    )
    return path in __context__[CONTEXT_BASE][saltenv]["dir_list"]


def path_exists(path, saltenv="base"):
    """
    Return ``True`` if a path exists in the state tree, ``False`` otherwise. The path
    could refer to a file or directory.

    .. versionadded:: 3004

    :param str path: The fully qualified path to a file or directory in the state tree.
    :param str saltenv: The fileserver environment to search. Default: ``base``

    CLI Example:

    .. code-block:: bash

        salt '*' slsutil.path_exists nginx/defaults.yaml
    """

    return file_exists(path, saltenv) or dir_exists(path, saltenv)


def findup(startpath, filenames, saltenv="base"):
    """
    Find the first path matching a filename or list of filenames in a specified
    directory or the nearest ancestor directory. Returns the full path to the
    first file found.

    .. versionadded:: 3004

    :param str startpath: The fileserver path from which to begin the search.
        An empty string refers to the state tree root.
    :param filenames: A filename or list of filenames to search for. Searching for
        directory names is also supported.
    :param str saltenv: The fileserver environment to search. Default: ``base``

    Example: return the path to ``defaults.yaml``, walking up the tree from the
    state file currently being processed.

    .. code-block:: jinja

        {{ salt["slsutil.findup"](tplfile, "defaults.yaml") }}

    CLI Example:

    .. code-block:: bash

        salt '*' slsutil.findup formulas/shared/nginx map.jinja
    """

    # Normalize the path
    if startpath:
        startpath = posixpath.normpath(startpath)

    # Verify the cwd is a valid path in the state tree
    if startpath and not path_exists(startpath, saltenv):
        raise salt.exceptions.SaltInvocationError(
            f"Starting path not found in the state tree: {startpath}"
        )

    # Ensure that patterns is a string or list of strings
    if isinstance(filenames, str):
        filenames = [filenames]
    if not isinstance(filenames, list):
        raise salt.exceptions.SaltInvocationError(
            "Filenames argument must be a string or list of strings"
        )

    while True:

        # Loop over filenames, looking for one at the current path level
        for filename in filenames:
            fullname = salt.utils.path.join(
                startpath or "", filename, use_posixpath=True
            )
            if path_exists(fullname, saltenv):
                return fullname

        # If the root path was just checked, raise an error
        if not startpath:
            raise salt.exceptions.CommandExecutionError(
                "File pattern(s) not found in path ancestry"
            )

        # Move up one level in the ancestry
        startpath = os.path.dirname(startpath)



"""
Decrypt pillar data through the builtin NACL renderer

In most cases, you'll want to make this the last external pillar used. For
example, to pair with the builtin stack pillar you could do something like
this:

.. code:: yaml

    nacl.config:
        keyfile: /root/.nacl

    ext_pillar:
      - stack: /path/to/stack.cfg
      - nacl: {}

Set ``nacl.config`` in your config.

"""

import salt

__virtualname__ = "nacl"


def __virtual__():
    if __opts__["fips_mode"] is True:
        return False, "nacl pillar data not available in FIPS mode"
    return __virtualname__


def ext_pillar(minion_id, pillar, *args, **kwargs):
    render_function = salt.loader.render(__opts__, __salt__).get("nacl")
    return render_function(pillar)



"""
Proxy Minion interface module for managing VMWare vCenters.

:codeauthor: `Rod McKenzie (roderick.mckenzie@morganstanley.com)`
:codeauthor: `Alexandru Bleotu (alexandru.bleotu@morganstanley.com)`

Dependencies
============

- pyVmomi Python Module

pyVmomi
-------

PyVmomi can be installed via pip:

.. code-block:: bash

    pip install pyVmomi

.. note::

    Version 6.0 of pyVmomi has some problems with SSL error handling on certain
    versions of Python. If using version 6.0 of pyVmomi, Python 2.6,
    Python 2.7.9, or newer must be present. This is due to an upstream dependency
    in pyVmomi 6.0 that is not supported in Python versions 2.7 to 2.7.8. If the
    version of Python is not in the supported range, you will need to install an
    earlier version of pyVmomi. See `Issue #29537`_ for more information.

.. _Issue #29537: https://github.com/saltstack/salt/issues/29537

Based on the note above, to install an earlier version of pyVmomi than the
version currently listed in PyPi, run the following:

.. code-block:: bash

    pip install pyVmomi==5.5.0.2014.1.1

The 5.5.0.2014.1.1 is a known stable version that this original ESXi State
Module was developed against.


Configuration
=============
To use this proxy module, please use on of the following configurations:


.. code-block:: yaml

    proxy:
      proxytype: vcenter
      vcenter: <ip or dns name of parent vcenter>
      username: <vCenter username>
      mechanism: userpass
      passwords:
        - first_password
        - second_password
        - third_password

    proxy:
      proxytype: vcenter
      vcenter: <ip or dns name of parent vcenter>
      username: <vCenter username>
      domain: <user domain>
      mechanism: sspi
      principal: <host kerberos principal>

proxytype
---------
The ``proxytype`` key and value pair is critical, as it tells Salt which
interface to load from the ``proxy`` directory in Salt's install hierarchy,
or from ``/srv/salt/_proxy`` on the Salt Master (if you have created your
own proxy module, for example). To use this Proxy Module, set this to
``vcenter``.

vcenter
-------
The location of the VMware vCenter server (host of ip). Required

username
--------
The username used to login to the vcenter, such as ``root``.
Required only for userpass.

mechanism
---------
The mechanism used to connect to the vCenter server. Supported values are
``userpass`` and ``sspi``. Required.

passwords
---------
A list of passwords to be used to try and login to the vCenter server. At least
one password in this list is required if mechanism is ``userpass``

The proxy integration will try the passwords listed in order.

domain
------
User domain. Required if mechanism is ``sspi``

principal
---------
Kerberos principal. Rquired if mechanism is ``sspi``

protocol
--------
If the vCenter is not using the default protocol, set this value to an
alternate protocol. Default is ``https``.

port
----
If the ESXi host is not using the default port, set this value to an
alternate port. Default is ``443``.


Salt Proxy
==========

After your pillar is in place, you can test the proxy. The proxy can run on
any machine that has network connectivity to your Salt Master and to the
vCenter server in the pillar. SaltStack recommends that the machine running the
salt-proxy process also run a regular minion, though it is not strictly
necessary.

On the machine that will run the proxy, make sure there is an ``/etc/salt/proxy``
file with at least the following in it:

.. code-block:: yaml

    master: <ip or hostname of salt-master>

You can then start the salt-proxy process with:

.. code-block:: bash

    salt-proxy --proxyid <id of the cluster>

You may want to add ``-l debug`` to run the above in the foreground in
debug mode just to make sure everything is OK.

Next, accept the key for the proxy on your salt-master, just like you
would for a regular minion:

.. code-block:: bash

    salt-key -a <id you gave the vcenter host>

You can confirm that the pillar data is in place for the proxy:

.. code-block:: bash

    salt <id> pillar.items

And now you should be able to ping the ESXi host to make sure it is
responding:

.. code-block:: bash

    salt <id> test.ping

At this point you can execute one-off commands against the vcenter. For
example, you can get if the proxy can actually connect to the vCenter:

.. code-block:: bash

    salt <id> vsphere.test_vcenter_connection

Note that you don't need to provide credentials or an ip/hostname. Salt
knows to use the credentials you stored in Pillar.

It's important to understand how this particular proxy works.
:mod:`Salt.modules.vsphere </ref/modules/all/salt.modules.vsphere>` is a
standard Salt execution module.

If you pull up the docs for it you'll see that almost every function in the
module takes credentials and a targets either a vcenter or a host. When
credentials and a host aren't passed, Salt runs commands through ``pyVmomi``
against the local machine. If you wanted, you could run functions from this
module on any host where an appropriate version of ``pyVmomi`` is installed,
and that host would reach out over the network and communicate with the ESXi
host.
"""

import logging
import os

import salt.exceptions
from salt.config.schemas.vcenter import VCenterProxySchema
from salt.utils.dictupdate import merge

# This must be present or the Salt loader won't load this module.
__proxyenabled__ = ["vcenter"]

# External libraries
try:
    import jsonschema

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False


# Variables are scoped to this module so we can have persistent data
# across calls to fns in here.
DETAILS = {}


# Set up logging
log = logging.getLogger(__name__)
# Define the module's virtual name
__virtualname__ = "vcenter"


def __virtual__():
    """
    Only load if the vsphere execution module is available.
    """
    if HAS_JSONSCHEMA:
        return __virtualname__

    return False, "The vcenter proxy module did not load."


def init(opts):
    """
    This function gets called when the proxy starts up.
    For login the protocol and port are cached.
    """
    log.info("Initting vcenter proxy module in process %s", os.getpid())
    log.trace("VCenter Proxy Validating vcenter proxy input")
    schema = VCenterProxySchema.serialize()
    log.trace("schema = %s", schema)
    proxy_conf = merge(opts.get("proxy", {}), __pillar__.get("proxy", {}))
    log.trace("proxy_conf = %s", proxy_conf)
    try:
        jsonschema.validate(proxy_conf, schema)
    except jsonschema.exceptions.ValidationError as exc:
        raise salt.exceptions.InvalidConfigError(exc)

    # Save mandatory fields in cache
    for key in ("vcenter", "mechanism"):
        DETAILS[key] = proxy_conf[key]

    # Additional validation
    if DETAILS["mechanism"] == "userpass":
        if "username" not in proxy_conf:
            raise salt.exceptions.InvalidConfigError(
                "Mechanism is set to 'userpass' , but no "
                "'username' key found in proxy config"
            )
        if "passwords" not in proxy_conf:
            raise salt.exceptions.InvalidConfigError(
                "Mechanism is set to 'userpass' , but no "
                "'passwords' key found in proxy config"
            )
        for key in ("username", "passwords"):
            DETAILS[key] = proxy_conf[key]
    else:
        if "domain" not in proxy_conf:
            raise salt.exceptions.InvalidConfigError(
                "Mechanism is set to 'sspi' , but no 'domain' key found in proxy config"
            )
        if "principal" not in proxy_conf:
            raise salt.exceptions.InvalidConfigError(
                "Mechanism is set to 'sspi' , but no "
                "'principal' key found in proxy config"
            )
        for key in ("domain", "principal"):
            DETAILS[key] = proxy_conf[key]

    # Save optional
    DETAILS["protocol"] = proxy_conf.get("protocol")
    DETAILS["port"] = proxy_conf.get("port")
    DETAILS["verify_ssl"] = proxy_conf.get("verify_ssl")
    DETAILS["ca_bundle"] = proxy_conf.get("ca_bundle")

    # Test connection
    if DETAILS["mechanism"] == "userpass":
        # Get the correct login details
        log.info(
            "Retrieving credentials and testing vCenter connection for "
            "mehchanism 'userpass'"
        )
        try:
            username, password = find_credentials()
        except salt.exceptions.SaltSystemExit as err:
            log.critical("Error: %s", err)
            return False
        else:
            DETAILS["password"] = password
    return True


def ping():
    """
    Returns True.

    CLI Example:

    .. code-block:: bash

        salt vcenter test.ping
    """
    return True


def shutdown():
    """
    Shutdown the connection to the proxy device. For this proxy,
    shutdown is a no-op.
    """
    log.debug("VCenter proxy shutdown() called...")


def find_credentials():
    """
    Cycle through all the possible credentials and return the first one that
    works.
    """

    # if the username and password were already found don't fo though the
    # connection process again
    if "username" in DETAILS and "password" in DETAILS:
        return DETAILS["username"], DETAILS["password"]

    passwords = __pillar__["proxy"]["passwords"]
    for password in passwords:
        DETAILS["password"] = password
        if not __salt__["vsphere.test_vcenter_connection"]():
            # We are unable to authenticate
            continue
        # If we have data returned from above, we've successfully authenticated.
        return DETAILS["username"], password
    # We've reached the end of the list without successfully authenticating.
    raise salt.exceptions.VMwareConnectionError(
        "Cannot complete login due to incorrect credentials."
    )


def get_details():
    """
    Function that returns the cached details
    """
    return DETAILS



"""
tests.pytests.unit.test_config
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Unit tests for salt's config modulet
"""

import pathlib

import salt.config
import salt.syspaths


def test_call_id_function(tmp_path):
    "Defining id_function works as expected"
    cache_dir = tmp_path / "cache"
    extmods = tmp_path / "extmods"
    opts = {
        "id_function": {"grains.get": {"key": "osfinger"}},
        "cachedir": str(cache_dir),
        "extension_modules": str(extmods),
        "grains": {"osfinger": "meh"},
        "optimization_order": [0],
    }
    ret = salt.config.call_id_function(opts)
    assert ret == "meh"


def test_prepend_root_dir(tmp_path):
    root = tmp_path / "root"
    opts = {
        "root_dir": root,
        "foo": str(pathlib.Path(salt.syspaths.ROOT_DIR) / "var" / "foo"),
    }
    salt.config.prepend_root_dir(opts, ["foo"])
    assert opts["foo"] == str(root / "var" / "foo")



import copy
import os
from collections import OrderedDict

import pytest

import salt.modules.grains as grainsmod
import salt.utils.dictupdate as dictupdate
from salt.exceptions import SaltException
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS


@pytest.fixture
def configure_loader_modules():
    conf_file = os.path.join(RUNTIME_VARS.TMP, "__salt_test_grains")
    cachedir = os.path.join(RUNTIME_VARS.TMP, "__salt_test_grains_cache_dir")
    if not os.path.isdir(cachedir):
        os.makedirs(cachedir)
    return {
        grainsmod: {
            "__opts__": {"conf_file": conf_file, "cachedir": cachedir},
            "__salt__": {"saltutil.refresh_grains": MagicMock()},
        }
    }


def test_filter_by():
    with patch.dict(
        grainsmod.__grains__,
        {"os_family": "MockedOS", "1": "1", "2": "2", "roles": ["A", "B"]},
    ):

        dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}
        dict2 = {
            "default": {"A": "B", "C": {"D": "E"}},
            "1": {"A": "X"},
            "2": {"C": {"D": "H"}},
            "MockedOS": {"A": "Z"},
        }

        mdict1 = {"D": {"E": "I"}, "J": "K"}
        mdict2 = {"A": "Z"}
        mdict3 = {"C": {"D": "J"}}

        # test None result with non existent grain and no default
        res = grainsmod.filter_by(dict1, grain="xxx")
        assert res is None

        # test None result with os_family grain and no matching result
        res = grainsmod.filter_by(dict1)
        assert res is None

        # test with non existent grain, and a given default key
        res = grainsmod.filter_by(dict1, grain="xxx", default="C")
        assert res == {"D": {"E": "F", "G": "H"}}

        # add a merge dictionary, F disappears
        res = grainsmod.filter_by(dict1, grain="xxx", merge=mdict1, default="C")
        assert res == {"D": {"E": "I", "G": "H"}, "J": "K"}
        # dict1 was altered, reestablish
        dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}

        # default is not present in dict1, check we only have merge in result
        res = grainsmod.filter_by(dict1, grain="xxx", merge=mdict1, default="Z")
        assert res == mdict1

        # default is not present in dict1, and no merge, should get None
        res = grainsmod.filter_by(dict1, grain="xxx", default="Z")
        assert res is None

        # test giving a list as merge argument raise exception
        pytest.raises(SaltException, grainsmod.filter_by, dict1, "xxx", ["foo"], "C")

        # Now, re-test with an existing grain (os_family), but with no match.
        res = grainsmod.filter_by(dict1)
        assert res is None
        res = grainsmod.filter_by(dict1, default="C")
        assert res == {"D": {"E": "F", "G": "H"}}
        res = grainsmod.filter_by(dict1, merge=mdict1, default="C")
        assert res == {"D": {"E": "I", "G": "H"}, "J": "K"}
        # dict1 was altered, reestablish
        dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1, merge=mdict1, default="Z")
        assert res == mdict1
        res = grainsmod.filter_by(dict1, default="Z")
        assert res is None
        # this one is in fact a traceback in updatedict, merging a string
        # with a dictionary
        pytest.raises(TypeError, grainsmod.filter_by, dict1, merge=mdict1, default="A")

        # Now, re-test with a matching grain.
        dict1 = {"A": "B", "MockedOS": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1)
        assert res == {"D": {"E": "F", "G": "H"}}
        res = grainsmod.filter_by(dict1, default="A")
        assert res == {"D": {"E": "F", "G": "H"}}
        res = grainsmod.filter_by(dict1, merge=mdict1, default="A")
        assert res == {"D": {"E": "I", "G": "H"}, "J": "K"}
        # dict1 was altered, reestablish
        dict1 = {"A": "B", "MockedOS": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1, merge=mdict1, default="Z")
        assert res == {"D": {"E": "I", "G": "H"}, "J": "K"}
        # dict1 was altered, reestablish
        dict1 = {"A": "B", "MockedOS": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1, default="Z")
        assert res == {"D": {"E": "F", "G": "H"}}

        # Test when grain value is a list
        dict1 = {"A": "B", "C": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1, grain="roles", default="C")
        assert res == "B"
        # Test default when grain value is a list
        dict1 = {"Z": "B", "C": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1, grain="roles", default="C")
        assert res == {"D": {"E": "F", "G": "H"}}

        # Test with wildcard pattern in the lookup_dict keys
        dict1 = {"*OS": "B", "C": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1)
        assert res == "B"
        # Test with non-strings in lookup_dict keys
        # Issue #38094
        dict1 = {1: 2, 3: {4: 5}, "*OS": "B"}
        res = grainsmod.filter_by(dict1)
        assert res == "B"
        # Test with sequence pattern with roles
        dict1 = {"Z": "B", "[BC]": {"D": {"E": "F", "G": "H"}}}
        res = grainsmod.filter_by(dict1, grain="roles", default="Z")
        assert res == {"D": {"E": "F", "G": "H"}}

        # Base tests
        # NOTE: these may fail to detect errors if dictupdate.update() is broken
        # but then the unit test for dictupdate.update() should fail and expose
        # that.  The purpose of these tests is it validate the logic of how
        # in filter_by() processes its arguments.

        # Test with just the base
        res = grainsmod.filter_by(dict2, grain="xxx", default="xxx", base="default")
        assert res == dict2["default"]
        # Test the base with the OS grain look-up
        res = grainsmod.filter_by(dict2, default="xxx", base="default")
        assert res == dictupdate.update(
            copy.deepcopy(dict2["default"]), dict2["MockedOS"]
        )
        # Test the base with default
        res = grainsmod.filter_by(dict2, grain="xxx", base="default")
        assert res == dict2["default"]

        res = grainsmod.filter_by(dict2, grain="1", base="default")
        assert res == dictupdate.update(copy.deepcopy(dict2["default"]), dict2["1"])

        res = grainsmod.filter_by(dict2, base="default", merge=mdict2)
        assert res == dictupdate.update(
            dictupdate.update(copy.deepcopy(dict2["default"]), dict2["MockedOS"]),
            mdict2,
        )

        res = grainsmod.filter_by(dict2, base="default", merge=mdict3)
        assert res == dictupdate.update(
            dictupdate.update(copy.deepcopy(dict2["default"]), dict2["MockedOS"]),
            mdict3,
        )


def test_append_not_a_list():
    with patch.dict(grainsmod.__grains__, {"b": "bval"}):
        res = grainsmod.append("b", "d")
        assert res == "The key b is not a valid list"
        assert grainsmod.__grains__ == {"b": "bval"}

        # Failing append to an existing dict
        with patch.dict(grainsmod.__grains__, {"b": {"b1": "bval1"}}):
            res = grainsmod.append("b", "d")
            assert res == "The key b is not a valid list"
            assert grainsmod.__grains__ == {"b": {"b1": "bval1"}}


def test_append_already_in_list():
    with patch.dict(grainsmod.__grains__, {"a_list": ["a", "b", "c"], "b": "bval"}):
        res = grainsmod.append("a_list", "b")
        assert res == "The val b was already in the list a_list"
        assert grainsmod.__grains__ == {"a_list": ["a", "b", "c"], "b": "bval"}


def test_append_ok():
    with patch.dict(grainsmod.__grains__, {"a_list": ["a", "b", "c"], "b": "bval"}):
        res = grainsmod.append("a_list", "d")
        assert res == {"a_list": ["a", "b", "c", "d"]}
        assert grainsmod.__grains__ == {"a_list": ["a", "b", "c", "d"], "b": "bval"}

    with patch.dict(grainsmod.__grains__, {"b": "bval"}):
        res = grainsmod.append("a_list", "d")
        assert res == {"a_list": ["d"]}
        assert grainsmod.__grains__ == {"a_list": ["d"], "b": "bval"}

    with patch.dict(grainsmod.__grains__, {"b": "bval"}):
        res = grainsmod.append("b", "d", convert=True)
        assert res == {"b": ["bval", "d"]}
        assert grainsmod.__grains__ == {"b": ["bval", "d"]}

    with patch.dict(grainsmod.__grains__, {"b": {"b1": "bval1"}}):
        res = grainsmod.append("b", "d", convert=True)
        assert res == {"b": [{"b1": "bval1"}, "d"]}
        assert grainsmod.__grains__ == {"b": [{"b1": "bval1"}, "d"]}


def test_append_nested_not_a_list():
    with patch.dict(grainsmod.__grains__, {"a": {"b": "bval"}}):
        res = grainsmod.append("a:b", "d")
        assert res == "The key a:b is not a valid list"
        assert grainsmod.__grains__ == {"a": {"b": "bval"}}

    with patch.dict(grainsmod.__grains__, {"a": {"b": {"b1": "bval1"}}}):
        res = grainsmod.append("a:b", "d")
        assert res == "The key a:b is not a valid list"
        assert grainsmod.__grains__ == {"a": {"b": {"b1": "bval1"}}}


def test_append_nested_already_in_list():
    with patch.dict(
        grainsmod.__grains__, {"a": {"a_list": ["a", "b", "c"], "b": "bval"}}
    ):
        res = grainsmod.append("a:a_list", "b")
        assert res == "The val b was already in the list a:a_list"
        assert grainsmod.__grains__ == {"a": {"a_list": ["a", "b", "c"], "b": "bval"}}


def test_append_nested_ok():
    with patch.dict(
        grainsmod.__grains__, {"a": {"a_list": ["a", "b", "c"], "b": "bval"}}
    ):
        res = grainsmod.append("a:a_list", "d")
        assert res == {"a": {"a_list": ["a", "b", "c", "d"], "b": "bval"}}
        assert grainsmod.__grains__ == {
            "a": {"a_list": ["a", "b", "c", "d"], "b": "bval"}
        }

    with patch.dict(grainsmod.__grains__, {"a": {"b": "bval"}}):
        res = grainsmod.append("a:a_list", "d")
        assert res == {"a": {"a_list": ["d"], "b": "bval"}}
        assert grainsmod.__grains__ == {"a": {"a_list": ["d"], "b": "bval"}}

    with patch.dict(grainsmod.__grains__, {"a": {"b": "bval"}}):
        res = grainsmod.append("a:b", "d", convert=True)
        assert res == {"a": {"b": ["bval", "d"]}}
        assert grainsmod.__grains__ == {"a": {"b": ["bval", "d"]}}

    with patch.dict(grainsmod.__grains__, {"a": {"b": {"b1": "bval1"}}}):
        res = grainsmod.append("a:b", "d", convert=True)
        assert res == {"a": {"b": [{"b1": "bval1"}, "d"]}}
        assert grainsmod.__grains__ == {"a": {"b": [{"b1": "bval1"}, "d"]}}


def test_append_to_an_element_of_a_list():
    with patch.dict(grainsmod.__grains__, {"a": ["b", "c"]}):
        res = grainsmod.append("a:b", "d")
        assert res == {"a": ["b", "c"]}
        assert grainsmod.__grains__ == {"a": ["b", "c"]}


def test_set_value_already_set():
    with patch.dict(grainsmod.__grains__, {"a": 12, "c": 8}):
        res = grainsmod.set("a", 12)
        assert res["result"]
        assert res["comment"] == "Grain is already set"
        assert grainsmod.__grains__ == {"a": 12, "c": 8}

    with patch.dict(grainsmod.__grains__, {"a": ["item", 12], "c": 8}):
        res = grainsmod.set("a", ["item", 12])
        assert res["result"]
        assert res["comment"] == "Grain is already set"
        assert grainsmod.__grains__ == {"a": ["item", 12], "c": 8}

    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.set("b,nested", "val", delimiter=",")
        assert res["result"]
        assert res["comment"] == "Grain is already set"
        assert grainsmod.__grains__ == {"a": "aval", "b": {"nested": "val"}, "c": 8}


def test_set_fail_replacing_existing_complex_key():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("a", ["item", 12])
        assert not res["result"]
        assert (
            res["comment"] == "The key 'a' exists and the given value is a "
            "dict or a list. Use 'force=True' to overwrite."
        )
        assert grainsmod.__grains__ == {"a": "aval", "c": 8}

    with patch.dict(grainsmod.__grains__, {"a": ["item", 12], "c": 8}):
        res = grainsmod.set("a", ["item", 14])
        assert not res["result"]
        assert (
            res["comment"] == "The key 'a' exists but is a dict or a list. "
            "Use 'force=True' to overwrite."
        )
        assert grainsmod.__grains__ == {"a": ["item", 12], "c": 8}

    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": ["val1"]}], "c": 8}
    ):
        res = grainsmod.set("b,l2", "val2", delimiter=",")
        assert not res["result"]
        assert (
            res["comment"] == "The key 'b,l2' exists but is a dict or a "
            "list. Use 'force=True' to overwrite."
        )
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": ["l1", {"l2": ["val1"]}],
            "c": 8,
        }


def test_set_nested_fails_replace_simple_value():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "b": "l1", "c": 8}):
        res = grainsmod.set("b,l3", "val3", delimiter=",")
        assert not res["result"]
        assert (
            res["comment"] == "The key 'b' value is 'l1', which is different from "
            "the provided key 'l3'. Use 'force=True' to overwrite."
        )
        assert grainsmod.__grains__ == {"a": "aval", "b": "l1", "c": 8}


def test_set_simple_value():
    with patch.dict(grainsmod.__grains__, {"a": ["b", "c"], "c": 8}):
        res = grainsmod.set("b", "bval")
        assert res["result"]
        assert res["changes"] == {"b": "bval"}
        assert grainsmod.__grains__ == {"a": ["b", "c"], "b": "bval", "c": 8}


def test_set_replace_value():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("a", 12)
        assert res["result"]
        assert res["changes"] == {"a": 12}
        assert grainsmod.__grains__ == {"a": 12, "c": 8}


def test_set_None_ok():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("b", None)
        assert res["result"]
        assert res["changes"] == {"b": None}
        assert grainsmod.__grains__ == {"a": "aval", "b": None, "c": 8}


def test_set_None_ok_destructive():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("b", None, destructive=True)
        assert res["result"]
        assert res["changes"] == {"b": None}
        assert grainsmod.__grains__ == {"a": "aval", "c": 8}


def test_set_None_replace_ok():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("a", None)
        assert res["result"]
        assert res["changes"] == {"a": None}
        assert grainsmod.__grains__ == {"a": None, "c": 8}


def test_set_None_force_destructive():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("a", None, force=True, destructive=True)
        assert res["result"]
        assert res["changes"] == {"a": None}
        assert grainsmod.__grains__ == {"c": 8}


def test_set_replace_value_was_complex_force():
    with patch.dict(grainsmod.__grains__, {"a": ["item", 12], "c": 8}):
        res = grainsmod.set("a", "aval", force=True)
        assert res["result"]
        assert res["changes"] == {"a": "aval"}
        assert grainsmod.__grains__ == {"a": "aval", "c": 8}


def test_set_complex_value_force():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("a", ["item", 12], force=True)
        assert res["result"]
        assert res["changes"] == {"a": ["item", 12]}
        assert grainsmod.__grains__ == {"a": ["item", 12], "c": 8}


def test_set_nested_create():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "c": 8}):
        res = grainsmod.set("b,nested", "val", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": {"nested": "val"}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {"nested": "val"}, "c": 8}


def test_set_nested_update_dict():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.set("b,nested", "val2", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": {"nested": "val2"}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {"nested": "val2"}, "c": 8}


def test_set_nested_update_dict_remove_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.set("b,nested", None, delimiter=",", destructive=True)
        assert res["result"]
        assert res["changes"] == {"b": {}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {}, "c": 8}


def test_set_nested_update_dict_new_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.set("b,b2", "val2", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": {"b2": "val2", "nested": "val"}}
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": {"b2": "val2", "nested": "val"},
            "c": 8,
        }


def test_set_nested_list_replace_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": ["l1", "l2", "l3"], "c": 8}
    ):
        res = grainsmod.set("b,l2", "val2", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": ["l1", {"l2": "val2"}, "l3"]}
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": ["l1", {"l2": "val2"}, "l3"],
            "c": 8,
        }


def test_set_nested_list_update_dict_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": "val1"}], "c": 8}
    ):
        res = grainsmod.set("b,l2", "val2", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": ["l1", {"l2": "val2"}]}
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": ["l1", {"l2": "val2"}],
            "c": 8,
        }


def test_set_nested_list_update_dict_key_overwrite():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": ["val1"]}], "c": 8}
    ):
        res = grainsmod.set("b,l2", "val2", delimiter=",", force=True)
        assert res["result"]
        assert res["changes"] == {"b": ["l1", {"l2": "val2"}]}
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": ["l1", {"l2": "val2"}],
            "c": 8,
        }


def test_set_nested_list_append_dict_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": ["l1", {"l2": "val2"}], "c": 8}
    ):
        res = grainsmod.set("b,l3", "val3", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": ["l1", {"l2": "val2"}, {"l3": "val3"}]}
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": ["l1", {"l2": "val2"}, {"l3": "val3"}],
            "c": 8,
        }


def test_set_nested_existing_value_is_the_key():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "b": "l3", "c": 8}):
        res = grainsmod.set("b,l3", "val3", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": {"l3": "val3"}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {"l3": "val3"}, "c": 8}


def test_set_nested_existing_value_overwrite():
    with patch.dict(grainsmod.__grains__, {"a": "aval", "b": "l1", "c": 8}):
        res = grainsmod.set("b,l3", "val3", delimiter=",", force=True)
        assert res["result"]
        assert res["changes"] == {"b": {"l3": "val3"}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {"l3": "val3"}, "c": 8}


def test_set_deeply_nested_update():
    with patch.dict(
        grainsmod.__grains__,
        {"a": "aval", "b": {"l1": ["l21", "l22", {"l23": "l23val"}]}, "c": 8},
    ):
        res = grainsmod.set("b,l1,l23", "val", delimiter=",")
        assert res["result"]
        assert res["changes"] == {"b": {"l1": ["l21", "l22", {"l23": "val"}]}}
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": {"l1": ["l21", "l22", {"l23": "val"}]},
            "c": 8,
        }


def test_set_deeply_nested_create():
    with patch.dict(
        grainsmod.__grains__,
        {"a": "aval", "b": {"l1": ["l21", "l22", {"l23": "l23val"}]}, "c": 8},
    ):
        res = grainsmod.set("b,l1,l24,l241", "val", delimiter=",")
        assert res["result"]
        assert res["changes"] == {
            "b": {
                "l1": [
                    "l21",
                    "l22",
                    {"l23": "l23val"},
                    {"l24": {"l241": "val"}},
                ]
            }
        }
        assert grainsmod.__grains__ == {
            "a": "aval",
            "b": {
                "l1": [
                    "l21",
                    "l22",
                    {"l23": "l23val"},
                    {"l24": {"l241": "val"}},
                ]
            },
            "c": 8,
        }


def test_get_ordered():
    with patch.dict(
        grainsmod.__grains__,
        OrderedDict(
            [
                ("a", "aval"),
                (
                    "b",
                    OrderedDict(
                        [
                            ("z", "zval"),
                            (
                                "l1",
                                ["l21", "l22", OrderedDict([("l23", "l23val")])],
                            ),
                        ]
                    ),
                ),
                ("c", 8),
            ]
        ),
    ):
        res = grainsmod.get("b")
        assert type(res) == OrderedDict
        # Check that order really matters
        assert res == OrderedDict(
            [
                ("z", "zval"),
                ("l1", ["l21", "l22", OrderedDict([("l23", "l23val")])]),
            ]
        )
        assert not res == OrderedDict(
            [
                ("l1", ["l21", "l22", OrderedDict([("l23", "l23val")])]),
                ("z", "zval"),
            ]
        )


def test_get_unordered():
    with patch.dict(
        grainsmod.__grains__,
        OrderedDict(
            [
                ("a", "aval"),
                (
                    "b",
                    OrderedDict(
                        [
                            ("z", "zval"),
                            (
                                "l1",
                                ["l21", "l22", OrderedDict([("l23", "l23val")])],
                            ),
                        ]
                    ),
                ),
                ("c", 8),
            ]
        ),
    ):
        res = grainsmod.get("b", ordered=False)
        assert type(res) == dict
        # Check that order doesn't matter
        assert res == OrderedDict(
            [
                ("l1", ["l21", "l22", OrderedDict([("l23", "l23val")])]),
                ("z", "zval"),
            ]
        )


def test_equals():
    with patch.dict(
        grainsmod.__grains__,
        OrderedDict(
            [
                ("a", "aval"),
                (
                    "b",
                    OrderedDict(
                        [
                            ("z", "zval"),
                            (
                                "l1",
                                ["l21", "l22", OrderedDict([("l23", "l23val")])],
                            ),
                        ]
                    ),
                ),
                ("c", 8),
            ]
        ),
    ):
        res = grainsmod.equals("a", "aval")
        assert type(res) == bool
        assert res
        res = grainsmod.equals("b:z", "zval")
        assert res
        res = grainsmod.equals("b:z", "aval")
        assert not res


def test_grains_setval_refresh_pillar():
    """
    Test that refresh_pillar kwarg is being passed correctly from grains.setval to saltutil.refresh_grains
    """
    ret = grainsmod.setval("test_grains_setval_refresh_pillar", "saltopotamus")
    grainsmod.__salt__["saltutil.refresh_grains"].assert_called_with(
        refresh_pillar=True
    )
    ret = grainsmod.setval(
        "test_grains_setval_refresh_pillar", "saltopotamus", refresh_pillar=True
    )
    grainsmod.__salt__["saltutil.refresh_grains"].assert_called_with(
        refresh_pillar=True
    )
    ret = grainsmod.setval(
        "test_grains_setval_refresh_pillar", "saltopotamus", refresh_pillar=False
    )
    grainsmod.__salt__["saltutil.refresh_grains"].assert_called_with(
        refresh_pillar=False
    )


def test_setval_unicode():
    key = "塩"  # salt
    value = "塩人生です"  # salt is life

    # Note: call setvals 2 times is important
    # 1: add key to conf grains
    # 2: update and read key from conf grains
    for _ in range(2):
        ret = grainsmod.setvals({key: value})
        assert key in ret
        assert ret[key] == value


def test_delval_single():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.delval("a")
        assert res["result"]
        assert res["changes"] == {"a": None}
        assert grainsmod.__grains__ == {"a": None, "b": {"nested": "val"}, "c": 8}


def test_delval_nested():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.delval("b:nested")
        assert res["result"]
        assert res["changes"] == {"b": {"nested": None}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {"nested": None}, "c": 8}


def test_delkey_single_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.delkey("a")
        assert res["result"]
        assert res["changes"] == {"a": None}
        assert grainsmod.__grains__ == {"b": {"nested": "val"}, "c": 8}


def test_delkey_nested_key():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.delkey("b:nested")
        assert res["result"]
        assert res["changes"] == {"b": {}}
        assert grainsmod.__grains__ == {"a": "aval", "b": {}, "c": 8}


def test_delkey_nested_key_force_needed():
    with patch.dict(
        grainsmod.__grains__, {"a": "aval", "b": {"nested": "val"}, "c": 8}
    ):
        res = grainsmod.delkey("b", force=True)
        assert res["comment"].find("Use 'force=True' to overwrite.") == -1
        assert res["result"]
        assert res["changes"] == {"b": None}
        assert grainsmod.__grains__ == {"a": "aval", "c": 8}



import logging

log = logging.getLogger(__name__)


def __virtual__():
    log.info("master tops test loaded")
    return "master_tops_test"


def top(**kwargs):
    log.info("master_tops_test")
    return {"base": ["master_tops_test"]}



"""
Tests for MasterMinion class
"""

import logging
import os
import pathlib

import pytest

import salt.minion

log = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def minion_config_overrides(master_opts):
    """Configure minion to use same root_dir and config path as master."""
    root_dir = pathlib.Path(master_opts["root_dir"])
    conf_file = root_dir / "conf" / "minion"
    yield {"conf_file": str(conf_file), "root_dir": str(root_dir)}


@pytest.fixture
def minion_d_include_value(minion_opts):
    """Create minion.d/test.conf with 'minion_d_value' config option."""
    conf_dir = pathlib.Path(minion_opts["conf_file"]).parent
    minion_include_dir = (conf_dir / minion_opts["default_include"]).parent
    test_conf = minion_include_dir / "test.conf"
    os.makedirs(minion_include_dir)
    with salt.utils.files.fopen(test_conf, "w") as test_conf:
        test_conf.write("minion_d_value: True")


def test_issue_64219_masterminion_no_minion_d_include(
    master_opts, minion_d_include_value
):
    """Create MasterMinion and test it doesn't get config from 'minion.d/*.conf'."""

    mminion = salt.minion.MasterMinion(master_opts)
    assert "minion_d_value" not in mminion.opts



"""
    :codeauthor: :email:`Carlos D. Álvaro <github@cdalvaro.io>`

    Unit tests for the Slack Webhook Returner.
"""

import pytest

import salt.returners.slack_webhook_return as slack_webhook
from tests.support.mock import patch


@pytest.fixture
def webhook():
    return "T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX"


@pytest.fixture
def author_icon():
    return "https://platform.slack-edge.com/img/default_application_icon.png"


@pytest.fixture
def show_tasks():
    return True


@pytest.fixture
def minion_name():
    return "MacPro"


@pytest.fixture
def ret(minion_name):
    return {
        "fun_args": ["config.vim"],
        "jid": "20181227105933129338",
        "return": {
            "file_|-vim files present_|-/Users/cdalvaro/_|-recurse": {
                "comment": "The directory /Users/cdalvaro/ is in the correct state",
                "changes": {},
                "name": "/Users/cdalvaro/",
                "start_time": "10:59:52.252830",
                "result": True,
                "duration": 373.25,
                "__run_num__": 3,
                "__sls__": "config.vim",
                "__id__": "vim files present",
            },
            "pkg_|-vim present_|-vim_|-installed": {
                "comment": "All specified packages are already installed",
                "name": "vim",
                "start_time": "10:59:36.830591",
                "result": True,
                "duration": 1280.127,
                "__run_num__": 0,
                "__sls__": "config.vim",
                "changes": {},
                "__id__": "vim present",
            },
            "git_|-salt vim plugin updated_|-https://github.com/saltstack/salt-vim.git_|-latest": {
                "comment": (
                    "https://github.com/saltstack/salt-vim.git cloned to"
                    " /Users/cdalvaro/.vim/pack/git-plugins/start/salt"
                ),
                "name": "https://github.com/saltstack/salt-vim.git",
                "start_time": "11:00:01.892757",
                "result": True,
                "duration": 11243.445,
                "__run_num__": 6,
                "__sls__": "config.vim",
                "changes": {
                    "new": (
                        "https://github.com/saltstack/salt-vim.git =>"
                        " /Users/cdalvaro/.vim/pack/git-plugins/start/salt"
                    ),
                    "revision": {
                        "new": "6ca9e3500cc39dd417b411435d58a1b720b331cc",
                        "old": None,
                    },
                },
                "__id__": "salt vim plugin updated",
            },
            "pkg_|-macvim present_|-caskroom/cask/macvim_|-installed": {
                "comment": (
                    "The following packages failed to install/update:"
                    " caskroom/cask/macvim"
                ),
                "name": "caskroom/cask/macvim",
                "start_time": "10:59:38.111119",
                "result": False,
                "duration": 14135.45,
                "__run_num__": 1,
                "__sls__": "config.vim",
                "changes": {},
                "__id__": "macvim present",
            },
        },
        "retcode": 2,
        "success": True,
        "fun": "state.apply",
        "id": minion_name,
        "out": "highstate",
    }


@pytest.fixture
def evnt_ret(minion_name):
    return [
        {
            "data": {
                "fun_args": ["config.vim"],
                "jid": "20181227105933129338",
                "return": {
                    "file_|-vim files present_|-/Users/cdalvaro/_|-recurse": {
                        "comment": "The directory /Users/cdalvaro/ is in the correct state",
                        "changes": {},
                        "name": "/Users/cdalvaro/",
                        "start_time": "10:59:52.252830",
                        "result": True,
                        "duration": 373.25,
                        "__run_num__": 3,
                        "__sls__": "config.vim",
                        "__id__": "vim files present",
                    },
                    "pkg_|-vim present_|-v ju im_|-installed": {
                        "comment": "All specified packages are already installed",
                        "name": "vim",
                        "start_time": "10:59:36.830591",
                        "result": True,
                        "duration": 1280.127,
                        "__run_num__": 0,
                        "__sls__": "config.vim",
                        "changes": {},
                        "__id__": "vim present",
                    },
                    "git_|-salt vim plugin updated_|-https://github.com/saltstack/salt-vim.git_|-latest": {
                        "comment": "https://github.com/saltstack/salt-vim.git cloned to /Users/cdalvaro/.vim/pack/git-plugins/start/salt",
                        "name": "https://github.com/saltstack/salt-vim.git",
                        "start_time": "11:00:01.892757",
                        "result": True,
                        "duration": 11243.445,
                        "__run_num__": 6,
                        "__sls__": "config.vim",
                        "changes": {
                            "new": "https://github.com/saltstack/salt-vim.git => /Users/cdalvaro/.vim/pack/git-plugins/start/salt",
                            "revision": {
                                "new": "6ca9e3500cc39dd417b411435d58a1b720b331cc",
                                "old": None,
                            },
                        },
                        "__id__": "salt vim plugin updated",
                    },
                    "pkg_|-macvim present_|-caskroom/cask/macvim_|-installed": {
                        "comment": "The following packages failed to install/update: caskroom/cask/macvim",
                        "name": "caskroom/cask/macvim",
                        "start_time": "10:59:38.111119",
                        "result": False,
                        "duration": 14135.45,
                        "__run_num__": 1,
                        "__sls__": "config.vim",
                        "changes": {},
                        "__id__": "macvim present",
                    },
                },
                "retcode": 2,
                "success": True,
                "fun": "state.apply",
                "id": minion_name,
                "out": "highstate",
            }
        }
    ]


@pytest.fixture
def expected_payload(minion_name, author_icon):
    return {
        "attachments": [
            {
                "title": "Success: False",
                "color": "#272727",
                "text": (
                    "Function: state.apply\nFunction Args: ['config.vim']\nJID:"
                    " 20181227105933129338\nTotal: 4\nDuration: 27.03 secs"
                ),
                "author_link": f"{minion_name}",
                "author_name": f"{minion_name}",
                "fallback": f"{minion_name} | Failed",
                "author_icon": author_icon,
            },
            {"color": "good", "title": "Unchanged: 2"},
            {
                "color": "warning",
                "fields": [
                    {
                        "short": False,
                        "value": "config.vim.sls | salt vim plugin updated",
                    }
                ],
                "title": "Changed: 1",
            },
            {
                "color": "danger",
                "fields": [
                    {"short": False, "value": "config.vim.sls | macvim present"}
                ],
                "title": "Failed: 1",
            },
        ]
    }


@pytest.fixture
def configure_loader_modules(webhook, author_icon, show_tasks):
    return {
        slack_webhook: {
            "__opts__": {
                "slack_webhook.webhook": webhook,
                "slack_webhook.author_icon": author_icon,
                "slack_webhook.success_title": "{id} | Succeeded",
                "slack_webhook.failure_title": "{id} | Failed",
                "slack_webhook.show_tasks": show_tasks,
            }
        }
    }


def test_no_webhook(ret):
    """
    Test returner stops if no webhook is defined
    """
    with patch.dict(slack_webhook.__opts__, {"slack_webhook.webhook": ""}):
        assert slack_webhook.returner(ret) is None


def test_returner(ret):
    """
    Test to see if the Slack Webhook returner sends a message
    """
    query_ret = {"body": "ok", "status": 200}
    with patch("salt.utils.http.query", return_value=query_ret):
        assert slack_webhook.returner(ret)


def test_event_return(evnt_ret):
    """
    Test to see if the Slack Webhook event_return sends a message
    """
    query_ret = {"body": "ok", "status": 200}
    with patch("salt.utils.http.query", return_value=query_ret):
        assert slack_webhook.event_return(evnt_ret)


def test_generate_payload_for_state_apply(
    ret, minion_name, show_tasks, expected_payload, author_icon
):
    """
    Test _generate_payload private method
    """
    test_title = f"{minion_name} | Failed"
    test_report = slack_webhook._generate_report(ret, show_tasks)

    custom_grains = slack_webhook.__grains__
    custom_grains["id"] = minion_name
    custom_grains["localhost"] = minion_name

    with patch.dict(slack_webhook.__grains__, custom_grains):
        test_payload = slack_webhook._generate_payload(
            author_icon, test_title, test_report
        )

    assert test_payload == expected_payload


def test_generate_payload_for_test_ping(minion_name, author_icon, show_tasks):
    """
    Test _generate_payload private method
    """

    test_ping_ret = {
        "jid": "20200124204109195206",
        "return": True,
        "retcode": 0,
        "id": minion_name,
        "fun": "test.ping",
        "fun_args": [],
        "success": True,
    }
    expected_payload = {
        "attachments": [
            {
                "fallback": f"{minion_name} | Succeeded",
                "color": "#272727",
                "author_name": minion_name,
                "author_link": minion_name,
                "author_icon": author_icon,
                "title": "Success: True",
                "text": "Function: test.ping\nJID: 20200124204109195206\n",
            },
            {"color": "good", "title": "Return: True"},
        ]
    }

    test_title = f"{minion_name} | Succeeded"
    test_report = slack_webhook._generate_report(test_ping_ret, show_tasks)

    custom_grains = slack_webhook.__grains__
    custom_grains["id"] = minion_name
    custom_grains["localhost"] = minion_name

    with patch.dict(slack_webhook.__grains__, custom_grains):
        test_payload = slack_webhook._generate_payload(
            author_icon, test_title, test_report
        )

    assert test_payload == expected_payload



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import datetime
import logging
from collections import OrderedDict

import pytest

import salt.modules.schedule as schedule
from salt.utils.event import SaltEvent
from tests.support.mock import MagicMock, call, mock_open, patch

log = logging.getLogger(__name__)


@pytest.fixture
def job1():
    return {
        "function": "test.ping",
        "maxrunning": 1,
        "name": "job1",
        "jid_include": True,
        "enabled": True,
    }


@pytest.fixture
def configure_loader_modules(minion_opts):
    minion_opts["schedule"] = {}
    return {schedule: {"__opts__": minion_opts}}


# 'purge' function tests: 1
@pytest.mark.slow_test
def test_purge(job1):
    """
    Test if it purge all the jobs currently scheduled on the minion.
    """
    _schedule_data = {"job1": job1}
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {}})
    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value={"complete": True, "schedule": {}}
    )
    patch_schedule_list = patch.object(
        schedule, "list_", MagicMock(return_value=_schedule_data)
    )

    with (
        patch_makedirs
    ), (
        patch_schedule_opts
    ), patch_schedule_event_fire, patch_schedule_get_event, patch_schedule_list:
        assert schedule.purge() == {
            "comment": ["Deleted job: job1 from schedule."],
            "changes": {"job1": "removed"},
            "result": True,
        }

    _schedule_data = {"job1": job1, "job2": job1, "job3": job1}
    comm = [
        "Deleted job: job1 from schedule.",
        "Deleted job: job2 from schedule.",
        "Deleted job: job3 from schedule.",
    ]

    changes = {"job1": "removed", "job2": "removed", "job3": "removed"}

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": "salt"}})
    patch_schedule_list = patch.object(
        schedule, "list_", MagicMock(return_value=_schedule_data)
    )

    with patch_makedirs, patch_schedule_opts, patch_schedule_list:
        schedule_config_file = schedule._get_schedule_config_file()
        with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
            ret = schedule.purge(offline=True)
            assert any([True for item in comm if item in ret["comment"]])
            assert ret["changes"] == changes
            assert ret["result"]

            _call = call(b"schedule: {}\n")
            write_calls = fopen_mock.filehandles[schedule_config_file][
                0
            ].write._mock_mock_calls
            assert _call in write_calls


# 'delete' function tests: 1
@pytest.mark.slow_test
def test_delete(job1):
    """
    Test if it delete a job from the minion's schedule.
    """
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {}})
    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value={"complete": True, "schedule": {}}
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        assert schedule.delete("job1") == {
            "comment": "Job job1 does not exist.",
            "changes": {},
            "result": False,
        }

    _schedule_data = {"job1": job1}
    patch_schedule_list = patch.object(
        schedule, "list_", MagicMock(return_value=_schedule_data)
    )

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": "salt"}})

    comm = "Deleted Job job1 from schedule."
    changes = {"job1": "removed"}
    with patch_makedirs, patch_schedule_opts, patch_schedule_list:
        schedule_config_file = schedule._get_schedule_config_file()
        with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
            assert schedule.delete("job1", offline="True") == {
                "comment": comm,
                "changes": changes,
                "result": True,
            }

            _call = call(b"schedule: {}\n")
            write_calls = fopen_mock.filehandles[schedule_config_file][
                0
            ].write._mock_mock_calls
            assert _call in write_calls


# 'build_schedule_item' function tests: 1
def test_build_schedule_item():
    """
    Test if it build a schedule job.
    """
    comment = (
        'Unable to use "seconds", "minutes", "hours", '
        'or "days" with "when" or "cron" options.'
    )
    comment1 = 'Unable to use "when" and "cron" options together.  Ignoring.'
    with patch.dict(schedule.__opts__, {"job1": {}}):
        assert schedule.build_schedule_item("") == {
            "comment": "Job name is required.",
            "result": False,
        }

        assert schedule.build_schedule_item("job1", function="test.ping") == {
            "function": "test.ping",
            "maxrunning": 1,
            "name": "job1",
            "jid_include": True,
            "enabled": True,
        }

        assert schedule.build_schedule_item(
            "job1", function="test.ping", seconds=3600, when="2400"
        ) == {"comment": comment, "result": False}

        assert schedule.build_schedule_item(
            "job1", function="test.ping", when="2400", cron="2"
        ) == {"comment": comment1, "result": False}


# 'build_schedule_item_invalid_when' function tests: 1


def test_build_schedule_item_invalid_when():
    """
    Test if it build a schedule job.
    """
    comment = 'Schedule item garbage for "when" in invalid.'
    with patch.dict(schedule.__opts__, {"job1": {}}):
        assert schedule.build_schedule_item(
            "job1", function="test.ping", when="garbage"
        ) == {"comment": comment, "result": False}


def test_build_schedule_item_invalid_jobs_args():
    """
    Test failure if job_arg and job_kwargs are passed correctly
    """
    comment1 = "job_kwargs is not a dict. please correct and try again."
    comment2 = "job_args is not a list. please correct and try again."
    with patch.dict(schedule.__opts__, {"job1": {}}):
        assert schedule.build_schedule_item(
            "job1", function="test.args", job_kwargs=[{"key1": "value1"}]
        ) == {"comment": comment1, "result": False}
    with patch.dict(schedule.__opts__, {"job1": {}}):
        assert schedule.build_schedule_item(
            "job1", function="test.args", job_args={"positional"}
        ) == {"comment": comment2, "result": False}


def test_build_schedule_item_jid_include():
    """
    Test build_schedule_item when jid_include is passed and not passed
    """
    ret = schedule.build_schedule_item("job1", function="test.args", jid_include=False)
    assert ret == {
        "function": "test.args",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": False,
    }

    ret = schedule.build_schedule_item("job1", function="test.args", jid_include=True)
    assert ret == {
        "function": "test.args",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }

    ret = schedule.build_schedule_item("job1", function="test.args")
    assert ret == {
        "function": "test.args",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }


# 'add' function tests: 1


@pytest.mark.slow_test
def test_add():
    """
    Test if it add a job to the schedule.
    """
    comm1 = "Job job1 already exists in schedule."
    comm2 = (
        'Error: Unable to use "seconds", "minutes", "hours", '
        'or "days" with "when" or "cron" options.'
    )
    comm3 = 'Unable to use "when" and "cron" options together.  Ignoring.'
    comm4 = "Job: job2 would be added to schedule."

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": "salt"}})
    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value={"complete": True, "schedule": {}}
    )

    with patch_makedirs, patch_schedule_opts, patch_schedule_event_fire:

        _ret_value = {"complete": True, "schedule": {"job1": {"salt": "salt"}}}
        patch_schedule_get_event = patch.object(
            SaltEvent, "get_event", return_value=_ret_value
        )

        with patch_schedule_get_event:
            assert schedule.add("job1") == {
                "comment": comm1,
                "changes": {},
                "result": False,
            }

        _ret_value = {"complete": True, "schedule": {}}
        patch_schedule_get_event = patch.object(
            SaltEvent, "get_event", return_value=_ret_value
        )
        with patch_schedule_get_event:
            assert schedule.add(
                "job2", function="test.ping", seconds=3600, when="2400"
            ) == {"comment": comm2, "changes": {}, "result": False}

        _ret_value = {"complete": True, "schedule": {}}
        patch_schedule_get_event = patch.object(
            SaltEvent, "get_event", return_value=_ret_value
        )
        with patch_schedule_get_event:
            assert schedule.add(
                "job2", function="test.ping", when="2400", cron="2"
            ) == {"comment": comm3, "changes": {}, "result": False}

        _ret_value = {"complete": True, "schedule": {}}
        patch_schedule_get_event = patch.object(
            SaltEvent, "get_event", return_value=_ret_value
        )
        with patch_schedule_get_event:
            assert schedule.add("job2", function="test.ping", test=True) == {
                "comment": comm4,
                "changes": {},
                "result": True,
            }

    comm1 = "Added job: job3 to schedule."
    changes1 = {"job3": "added"}

    with patch_makedirs, patch_schedule_opts, patch_schedule_event_fire:
        schedule_config_file = schedule._get_schedule_config_file()
        with patch("os.path.exists", MagicMock(return_value=True)):
            with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
                assert schedule.add(
                    "job3", function="test.ping", seconds=3600, offline="True"
                ) == {"comment": comm1, "changes": changes1, "result": True}

                _call = call(
                    b"schedule:\n  job3: {function: test.ping, seconds: 3600, maxrunning: 1, name: job3, enabled: true,\n    jid_include: true}\n"
                )
                write_calls = fopen_mock.filehandles[schedule_config_file][
                    1
                ].write._mock_mock_calls
                assert _call in write_calls


# 'run_job' function tests: 1


@pytest.mark.slow_test
def test_run_job(job1):
    """
    Test if it run a scheduled job on the minion immediately.
    """
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})

    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent,
        "get_event",
        return_value={"complete": True, "schedule": {"job1": job1}},
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        assert schedule.run_job("job1") == {
            "comment": "Scheduling Job job1 on minion.",
            "result": True,
        }


# 'enable_job' function tests: 1


@pytest.mark.slow_test
def test_enable_job():
    """
    Test if it enable a job in the minion's schedule.
    """
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})

    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value={"complete": True, "schedule": {}}
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        assert schedule.enable_job("job1") == {
            "comment": "Job job1 does not exist.",
            "changes": {},
            "result": False,
        }


# 'disable_job' function tests: 1


@pytest.mark.slow_test
def test_disable_job():
    """
    Test if it disable a job in the minion's schedule.
    """
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})

    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value={"complete": True, "schedule": {}}
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        assert schedule.disable_job("job1") == {
            "comment": "Job job1 does not exist.",
            "changes": {},
            "result": False,
        }


# 'save' function tests: 1


@pytest.mark.slow_test
def test_save():
    """
    Test if it save all scheduled jobs on the minion.
    """
    comm1 = "Schedule (non-pillar items) saved."
    with patch.dict(
        schedule.__opts__,
        {"default_include": "/tmp"},
    ):
        with patch("os.makedirs", MagicMock(return_value=True)):
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                _ret_value = {"complete": True, "schedule": {}}
                with patch.object(SaltEvent, "get_event", return_value=_ret_value):
                    assert schedule.save() == {"comment": comm1, "result": True}


# 'enable' function tests: 1


def test_enable():
    """
    Test if it enable all scheduled jobs on the minion.
    """
    assert schedule.enable(test=True) == {
        "comment": "Schedule would be enabled.",
        "changes": {},
        "result": True,
    }


# 'disable' function tests: 1


def test_disable():
    """
    Test if it disable all scheduled jobs on the minion.
    """
    assert schedule.disable(test=True) == {
        "comment": "Schedule would be disabled.",
        "changes": {},
        "result": True,
    }


# 'move' function tests: 1


@pytest.mark.slow_test
def test_move(job1):
    """
    Test if it move scheduled job to another minion or minions.
    """
    comm1 = "no servers answered the published schedule.add command"
    comm2 = "the following minions return False"
    comm3 = "Moved Job job1 from schedule."

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})

    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent,
        "get_event",
        return_value={"complete": True, "schedule": {"job1": job1}},
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        mock = MagicMock(return_value={})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            assert schedule.move("job1", "minion1") == {
                "comment": comm1,
                "result": True,
            }

        mock = MagicMock(return_value={"minion1": ""})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            assert schedule.move("job1", "minion1") == {
                "comment": comm2,
                "minions": ["minion1"],
                "result": True,
            }

        mock = MagicMock(return_value={"minion1": "job1"})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                assert schedule.move("job1", "minion1") == {
                    "comment": comm3,
                    "minions": ["minion1"],
                    "result": True,
                }

        assert schedule.move("job3", "minion1") == {
            "comment": "Job job3 does not exist.",
            "result": False,
        }

        mock = MagicMock(side_effect=[{}, {"job1": {}}])

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        with patch.dict(schedule.__pillar__, {"schedule": {"job1": job1}}):
            mock = MagicMock(return_value={})
            patch_schedule_publish = patch.dict(
                schedule.__salt__, {"publish.publish": mock}
            )
            with patch_schedule_publish:
                assert schedule.move("job1", "minion1") == {
                    "comment": comm1,
                    "result": True,
                }

            mock = MagicMock(return_value={"minion1": ""})
            patch_schedule_publish = patch.dict(
                schedule.__salt__, {"publish.publish": mock}
            )
            with patch_schedule_publish:
                assert schedule.move("job1", "minion1") == {
                    "comment": comm2,
                    "minions": ["minion1"],
                    "result": True,
                }

            mock = MagicMock(return_value={"minion1": "job1"})
            patch_schedule_publish = patch.dict(
                schedule.__salt__, {"publish.publish": mock}
            )
            with patch_schedule_publish:
                assert schedule.move("job1", "minion1") == {
                    "comment": comm3,
                    "minions": ["minion1"],
                    "result": True,
                }


# 'copy' function tests: 1


@pytest.mark.slow_test
def test_copy(job1):
    """
    Test if it copy scheduled job to another minion or minions.
    """
    comm1 = "no servers answered the published schedule.add command"
    comm2 = "the following minions return False"
    comm3 = "Copied Job job1 from schedule to minion(s)."

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})

    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent,
        "get_event",
        return_value={"complete": True, "schedule": {"job1": job1}},
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        mock = MagicMock(return_value={})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch.dict(schedule.__salt__, {"publish.publish": mock}):
            assert schedule.copy("job1", "minion1") == {
                "comment": comm1,
                "result": True,
            }

        mock = MagicMock(return_value={"minion1": ""})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            assert schedule.copy("job1", "minion1") == {
                "comment": comm2,
                "minions": ["minion1"],
                "result": True,
            }

        mock = MagicMock(return_value={"minion1": "job1"})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                assert schedule.copy("job1", "minion1") == {
                    "comment": comm3,
                    "minions": ["minion1"],
                    "result": True,
                }

        assert schedule.copy("job3", "minion1") == {
            "comment": "Job job3 does not exist.",
            "result": False,
        }

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_event_fire, patch_schedule_get_event:
        mock = MagicMock(return_value={})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            assert schedule.copy("job1", "minion1") == {
                "comment": comm1,
                "result": True,
            }

        mock = MagicMock(return_value={"minion1": ""})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            assert schedule.copy("job1", "minion1") == {
                "comment": comm2,
                "minions": ["minion1"],
                "result": True,
            }

        mock = MagicMock(return_value={"minion1": "job1"})
        patch_schedule_publish = patch.dict(
            schedule.__salt__, {"publish.publish": mock}
        )
        with patch_schedule_publish:
            mock = MagicMock(return_value=True)
            with patch.dict(schedule.__salt__, {"event.fire": mock}):
                assert schedule.copy("job1", "minion1") == {
                    "comment": comm3,
                    "minions": ["minion1"],
                    "result": True,
                }


# 'modify' function tests: 1


@pytest.mark.slow_test
def test_modify(job1):
    """
    Test if modifying job to the schedule.
    """
    current_job1 = {
        "function": "salt",
        "seconds": "3600",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }

    new_job1 = {
        "function": "salt",
        "seconds": "60",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }

    comm1 = "Modified job: job1 in schedule."
    changes1 = {
        "job1": {
            "new": OrderedDict(new_job1),
            "old": OrderedDict(current_job1),
        }
    }

    new_job4 = {
        "function": "test.version",
        "seconds": "3600",
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }

    changes4 = {
        "job1": {
            "new": OrderedDict(new_job4),
            "old": OrderedDict(current_job1),
        }
    }

    expected1 = {"comment": comm1, "changes": changes1, "result": True}

    comm2 = (
        'Error: Unable to use "seconds", "minutes", "hours", '
        'or "days" with "when" option.'
    )
    expected2 = {"comment": comm2, "changes": {}, "result": False}

    comm3 = 'Unable to use "when" and "cron" options together.  Ignoring.'
    expected3 = {"comment": comm3, "changes": {}, "result": False}

    comm4 = "Job: job1 would be modified in schedule."
    expected4 = {"comment": comm4, "changes": changes4, "result": True}

    comm5 = "Job job2 does not exist in schedule."
    expected5 = {"comment": comm5, "changes": {}, "result": False}

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))

    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})

    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )

    with patch_makedirs, patch_schedule_opts, patch_schedule_event_fire:

        _ret_value = {"complete": True, "schedule": {"job1": current_job1}}
        patch_schedule_get_event = patch.object(
            SaltEvent, "get_event", return_value=_ret_value
        )
        with patch_schedule_get_event:
            ret = schedule.modify("job1", seconds="60")
            assert "job1" in ret["changes"]
            assert "new" in ret["changes"]["job1"]
            assert "old" in ret["changes"]["job1"]

            for key in [
                "maxrunning",
                "function",
                "seconds",
                "jid_include",
                "name",
                "enabled",
            ]:
                assert (
                    ret["changes"]["job1"]["new"][key]
                    == expected1["changes"]["job1"]["new"][key]
                )
                assert (
                    ret["changes"]["job1"]["old"][key]
                    == expected1["changes"]["job1"]["old"][key]
                )

            assert ret["comment"] == expected1["comment"]
            assert ret["result"] == expected1["result"]

            ret = schedule.modify(
                "job1", function="test.ping", seconds=3600, when="2400"
            )
            assert ret == expected2

            ret = schedule.modify("job1", function="test.ping", when="2400", cron="2")
            assert ret == expected3

            ret = schedule.modify("job1", function="test.version", test=True)

            assert "job1" in ret["changes"]
            assert "new" in ret["changes"]["job1"]
            assert "old" in ret["changes"]["job1"]

            for key in [
                "maxrunning",
                "function",
                "jid_include",
                "name",
                "enabled",
            ]:
                assert (
                    ret["changes"]["job1"]["new"][key]
                    == expected4["changes"]["job1"]["new"][key]
                )
                assert (
                    ret["changes"]["job1"]["old"][key]
                    == expected4["changes"]["job1"]["old"][key]
                )

            assert ret["comment"] == expected4["comment"]
            assert ret["result"] == expected4["result"]

        _ret_value = {"complete": True, "schedule": {}}
        patch_schedule_get_event = patch.object(
            SaltEvent, "get_event", return_value=_ret_value
        )
        with patch_schedule_get_event:
            ret = schedule.modify("job2", function="test.version", test=True)
            assert ret == expected5

    _schedule_data = {"job1": job1}
    comm = "Modified job: job1 in schedule."
    changes = {"job1": "removed"}

    changes = {
        "job1": {
            "new": OrderedDict(
                [
                    ("function", "test.version"),
                    ("maxrunning", 1),
                    ("name", "job1"),
                    ("enabled", True),
                    ("jid_include", True),
                ]
            ),
            "old": OrderedDict(
                [
                    ("function", "test.ping"),
                    ("maxrunning", 1),
                    ("name", "job1"),
                    ("jid_include", True),
                    ("enabled", True),
                ]
            ),
        }
    }
    patch_schedule_list = patch.object(
        schedule, "list_", MagicMock(return_value=_schedule_data)
    )

    with patch_makedirs, patch_schedule_opts, patch_schedule_list:
        schedule_config_file = schedule._get_schedule_config_file()
        with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
            ret = schedule.modify("job1", function="test.version", offline="True")
            assert ret["comment"] == comm
            assert ret["result"]
            assert all(
                [
                    True
                    for k, v in ret["changes"]["job1"]["old"].items()
                    if v == changes["job1"]["old"][k]
                ]
            )
            assert all(
                [
                    True
                    for k, v in ret["changes"]["job1"]["new"].items()
                    if v == changes["job1"]["new"][k]
                ]
            )

            _call = call(
                b"schedule:\n  job1: {enabled: true, function: test.version, jid_include: true, maxrunning: 1,\n    name: job1}\n"
            )
            write_calls = fopen_mock.filehandles[schedule_config_file][
                0
            ].write._mock_mock_calls
            assert _call in write_calls


# 'is_enabled' function tests: 1


def test_is_enabled():
    """
    Test is_enabled
    """
    job1 = {"function": "salt", "seconds": 3600}

    comm1 = "Modified job: job1 in schedule."

    mock_schedule = {"enabled": True, "job1": job1}

    mock_lst = MagicMock(return_value=mock_schedule)

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {}})
    patch_schedule_salt = patch.dict(
        schedule.__salt__,
        {"event.fire": MagicMock(return_value=True), "schedule.list": mock_lst},
    )
    patch_schedule_get_event = patch.object(
        SaltEvent,
        "get_event",
        return_value={"complete": True, "schedule": {"job1": job1}},
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_get_event, patch_schedule_salt:
        ret = schedule.is_enabled("job1")
        assert ret == job1

        ret = schedule.is_enabled()
        assert ret


# 'job_status' function tests: 1


def test_job_status():
    """
    Test is_enabled
    """
    job1 = {
        "_last_run": datetime.datetime(2021, 11, 1, 12, 36, 57),
        "_next_fire_time": datetime.datetime(2021, 11, 1, 13, 36, 57),
        "function": "salt",
        "seconds": 3600,
    }

    comm1 = "Modified job: job1 in schedule."

    mock_schedule = {"enabled": True, "job1": job1}

    mock_lst = MagicMock(return_value=mock_schedule)

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {"job1": job1}})
    patch_schedule_salt = patch.dict(
        schedule.__salt__,
        {"event.fire": MagicMock(return_value=True), "schedule.list": mock_lst},
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value={"complete": True, "data": job1}
    )

    with (
        patch_makedirs
    ), patch_schedule_opts, patch_schedule_get_event, patch_schedule_salt:
        ret = schedule.job_status("job1")
        assert ret == {
            "_last_run": "2021-11-01T12:36:57",
            "_next_fire_time": "2021-11-01T13:36:57",
            "function": "salt",
            "seconds": 3600,
        }


# 'purge' function tests: 1
@pytest.mark.slow_test
def test_list(job1):
    """
    Test schedule.list
    """
    _schedule_data = {"job1": job1}
    _ret_schedule_data = {
        "function": "test.ping",
        "seconds": 10,
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }
    _ret_value = {"complete": True, "schedule": {"job1": _ret_schedule_data}}
    saved_schedule = """schedule:
  job1: {enabled: true, function: test.ping, jid_include: true, maxrunning: 1, name: job1,
    seconds: 10}
"""
    expected = """schedule:
  job1:
    enabled: true
    function: test.ping
    jid_include: true
    maxrunning: 1
    name: job1
    saved: true
    seconds: 10
"""

    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {}})
    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value=_ret_value
    )
    patch_schedule_os_path_exists = patch(
        "os.path.exists", MagicMock(return_value=True)
    )

    with (
        patch_schedule_opts
    ), (
        patch_schedule_event_fire
    ), patch_schedule_get_event, patch_schedule_os_path_exists, patch_makedirs:
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=saved_schedule)
        ) as fopen_mock:
            ret = schedule.list_(offline=True)
            assert ret == expected

    _schedule_data = {"job1": job1}
    _ret_schedule_data = {
        "function": "test.ping",
        "seconds": 10,
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }
    _ret_value = {"complete": True, "schedule": {"job1": _ret_schedule_data}}
    expected = """schedule:
  job1:
    enabled: true
    function: test.ping
    jid_include: true
    maxrunning: 1
    name: job1
    saved: false
    seconds: 10
"""

    with (
        patch_schedule_opts
    ), (
        patch_makedirs
    ), (
        patch_schedule_event_fire
    ), patch_schedule_get_event, patch_schedule_os_path_exists:
        with patch("salt.utils.files.fopen", mock_open(read_data="")) as fopen_mock:
            ret = schedule.list_()
            assert ret == expected

    _ret_schedule_data = {
        "function": "test.ping",
        "seconds": 10,
        "maxrunning": 1,
        "name": "job1",
        "enabled": True,
        "jid_include": True,
    }

    _ret_value = {"complete": True, "schedule": {"job1": _ret_schedule_data}}

    saved_schedule = """schedule:
  job1: {enabled: true, function: test.ping, jid_include: true, maxrunning: 1, name: job1,
    seconds: 10}
"""

    expected = """schedule:
  job1:
    enabled: true
    function: test.ping
    jid_include: true
    maxrunning: 1
    name: job1
    saved: true
    seconds: 10
"""

    with (
        patch_schedule_opts
    ), (
        patch_schedule_event_fire
    ), patch_schedule_get_event, patch_schedule_os_path_exists, patch_makedirs:
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=saved_schedule)
        ) as fopen_mock:
            ret = schedule.list_()
            assert ret == expected


@pytest.mark.slow_test
def test_list_global_enabled(job1):
    """
    Test schedule.list when enabled globally
    """
    _schedule_data = {"enabled": True, "job1": job1}
    _ret_schedule_data = {
        "job1": {
            "function": "test.ping",
            "seconds": 10,
            "maxrunning": 1,
            "name": "job1",
            "enabled": True,
            "jid_include": True,
        },
        "enabled": True,
    }
    _ret_value = {"complete": True, "schedule": _ret_schedule_data}

    saved_schedule = """schedule:
  enabled: true
  job1: {enabled: true, function: test.ping, jid_include: true, maxrunning: 1, name: job1,
    seconds: 10}
"""

    expected = """schedule:
  enabled: true
  job1:
    enabled: true
    function: test.ping
    jid_include: true
    maxrunning: 1
    name: job1
    saved: true
    seconds: 10
"""
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {}})
    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value=_ret_value
    )
    patch_schedule_os_path_exists = patch(
        "os.path.exists", MagicMock(return_value=True)
    )

    with (
        patch_schedule_opts
    ), (
        patch_schedule_event_fire
    ), patch_schedule_get_event, patch_schedule_os_path_exists, patch_makedirs:
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=saved_schedule)
        ) as fopen_mock:
            ret = schedule.list_()
            assert ret == expected


@pytest.mark.slow_test
def test_list_global_disabled(job1):
    """
    Test schedule.list when disabled  globally
    """
    _schedule_data = {"disabled": True, "job1": job1}
    _ret_schedule_data = {
        "job1": {
            "function": "test.ping",
            "seconds": 10,
            "maxrunning": 1,
            "name": "job1",
            "enabled": True,
            "jid_include": True,
        },
        "enabled": False,
    }

    _ret_value = {"complete": True, "schedule": _ret_schedule_data}

    saved_schedule = """schedule:
  enabled: false
  job1: {enabled: true, function: test.ping, jid_include: true, maxrunning: 1, name: job1,
    seconds: 10}
"""

    expected = """schedule:
  enabled: false
  job1:
    enabled: true
    function: test.ping
    jid_include: true
    maxrunning: 1
    name: job1
    saved: true
    seconds: 10
"""
    patch_makedirs = patch("os.makedirs", MagicMock(return_value=True))
    patch_schedule_opts = patch.dict(schedule.__opts__, {"schedule": {}})
    patch_schedule_event_fire = patch.dict(
        schedule.__salt__, {"event.fire": MagicMock(return_value=True)}
    )
    patch_schedule_get_event = patch.object(
        SaltEvent, "get_event", return_value=_ret_value
    )
    patch_schedule_os_path_exists = patch(
        "os.path.exists", MagicMock(return_value=True)
    )

    with (
        patch_schedule_opts
    ), (
        patch_schedule_event_fire
    ), patch_schedule_get_event, patch_schedule_os_path_exists, patch_makedirs:
        with patch(
            "salt.utils.files.fopen", mock_open(read_data=saved_schedule)
        ) as fopen_mock:
            ret = schedule.list_()
            assert ret == expected



import time

import pytest

import salt.netapi.rest_tornado.saltnado as saltnado_app
from tests.support.mock import MagicMock, patch


@pytest.fixture
def arg_mock():
    mock = MagicMock()
    mock.opts = {
        "syndic_wait": 0.1,
        "cachedir": "/tmp/testing/cachedir",
        "sock_dir": "/tmp/testing/sock_drawer",
        "transport": "zeromq",
        "extension_modules": "/tmp/testing/moduuuuules",
        "order_masters": False,
        "gather_job_timeout": 10.001,
    }
    return mock


def test__verify_auth(arg_mock):
    base_handler = saltnado_app.BaseSaltAPIHandler(arg_mock, arg_mock)
    with patch.object(base_handler, "get_cookie", return_value="ABCDEF"):
        with patch.object(
            base_handler.application.auth,
            "get_tok",
            return_value={"expire": time.time() + 60},
        ):
            assert base_handler._verify_auth()


def test__verify_auth_expired(arg_mock):
    base_handler = saltnado_app.BaseSaltAPIHandler(arg_mock, arg_mock)
    with patch.object(base_handler, "get_cookie", return_value="ABCDEF"):
        with patch.object(
            base_handler.application.auth,
            "get_tok",
            return_value={"expire": time.time() - 60},
        ):
            assert not base_handler._verify_auth()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>

    Test cases for salt.modules.riak
"""

import pytest

import salt.modules.riak as riak
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {riak: {}}


def test_start():
    """
    Test for start Riak
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.start() == {"success": True, "comment": "success"}


def test_stop():
    """
    Test for stop Riak
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.stop() == {"success": True, "comment": "success"}


def test_cluster_join():
    """
    Test for Join a Riak cluster
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.cluster_join("A", "B") == {"success": True, "comment": "success"}


def test_cluster_leave():
    """
    Test for leaving a Riak cluster
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.cluster_leave("A", "B") == {"success": True, "comment": "success"}


def test_cluster_plan():
    """
    Test for Review Cluster Plan
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.cluster_plan()


def test_cluster_commit():
    """
    Test for Commit Cluster Changes
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.cluster_commit() == {"success": True, "comment": "success"}


def test_member_status():
    """
    Test for Get cluster member status
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"stdout": "A:a/B:b\nC:c/D:d"}
    ):
        assert riak.member_status() == {
            "membership": {},
            "summary": {
                "A": "a",
                "C": "c",
                "B": "b",
                "D": "d",
                "Exiting": 0,
                "Down": 0,
                "Valid": 0,
                "Leaving": 0,
                "Joining": 0,
            },
        }


def test_status():
    """
    Test status information
    """
    ret = {"stdout": "vnode_map_update_time_95 : 0\nvnode_map_update_time_99 : 0"}

    with patch.object(riak, "__execute_cmd", return_value=ret):
        assert riak.status() == {
            "vnode_map_update_time_95": "0",
            "vnode_map_update_time_99": "0",
        }


def test_test():
    """
    Test the Riak test
    """
    with patch.object(
        riak, "__execute_cmd", return_value={"retcode": 0, "stdout": "success"}
    ):
        assert riak.test() == {"success": True, "comment": "success"}


def test_services():
    """
    Test Riak Service List
    """
    with patch.object(riak, "__execute_cmd", return_value={"stdout": "[a,b,c]"}):
        assert riak.services() == ["a", "b", "c"]



"""
Test the salt.utils.path Salt Util
"""

import subprocess

import pytest

import salt.utils.path

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def symlink_file(tmp_path):
    target = tmp_path / "tgt_symlink"
    target.write_text("this is the symlink target")
    assert target.is_file()
    symlink = tmp_path / "symlink"
    symlink.symlink_to(target)
    assert symlink.is_symlink()
    assert symlink.read_text() == "this is the symlink target"
    assert symlink.is_file()
    yield symlink


@pytest.fixture
def symlink_dir(tmp_path):
    target = tmp_path / "tgt_symlink"
    target.mkdir()
    assert target.is_dir()
    symlink = tmp_path / "symlink"
    symlink.symlink_to(target)
    assert symlink.is_symlink()
    assert symlink.is_dir()
    yield symlink


@pytest.fixture
def junction(tmp_path):
    """
    Create a directory and a junction to that directory.
    """
    target = tmp_path / "tgt_junction"
    target.mkdir()
    assert target.is_dir()
    junction = tmp_path / "junction"
    cmd = ["cmd", "/c", "mklink", "/j", str(junction), str(target)]
    subprocess.run(
        cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, check=True
    )
    assert not junction.is_symlink()
    assert junction.is_dir()
    yield junction


def test_islink_symlink_file(symlink_file):
    assert salt.utils.path.islink(str(symlink_file))


def test_islink_symlink_dir(symlink_dir):
    assert salt.utils.path.islink(str(symlink_dir))


@pytest.mark.skip_unless_on_windows
def test_islink_junction(junction):
    """
    Junctions are only a thing on Windows, but they are the equivalent of a
    symlink. They just apply to directories
    """
    assert salt.utils.path.islink(str(junction))



from collections import OrderedDict

import pytest

import salt.states.smartos as smartos
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules():
    return {smartos: {"__opts__": {"test": False}}}


def test_config_present_does_not_exist():
    """
    Test salt.states.smartos.config_present
    when the config files does not exist
    """
    name = "test"
    value = "test_value"
    with patch("os.path.isfile", return_value=False):
        with patch("salt.utils.atomicfile.atomic_open", side_effect=IOError):
            ret = smartos.config_present(name=name, value=value)
    assert not ret["result"]
    assert (
        ret["comment"]
        == f'Could not add property {name} with value "{value}" to config'
    )


def test_parse_vmconfig_vrrp():
    """
    Test _parse_vmconfig's vrid -> mac convertor

    SmartOS will always use a mac based on the vrrp_vrid,
    so we will replace the provided mac with the one based
    on this value.

    Doing so ensures that 'old' nics are removed and 'new'
    nics get added as these actions are keyed on the mac
    property.
    """
    # NOTE: vmconfig is not a full vmadm payload,
    #       this is not an issue given we are only testing
    #       the vrrp_vrid to mac conversions
    ret = smartos._parse_vmconfig(
        OrderedDict(
            [
                (
                    "nics",
                    OrderedDict(
                        [
                            (
                                "00:00:5e:00:01:01",
                                OrderedDict(
                                    [
                                        ("vrrp_vrid", 1),
                                        ("vrrp_primary_ip", "12.34.5.6"),
                                    ]
                                ),
                            ),
                            (
                                "00:00:5e:00:01:24",
                                OrderedDict(
                                    [
                                        ("vrrp_vrid", 240),
                                        ("vrrp_primary_ip", "12.34.5.6"),
                                    ]
                                ),
                            ),
                            (
                                "00:22:06:00:00:01",
                                OrderedDict([("ips", ["12.34.5.6/24"])]),
                            ),
                        ]
                    ),
                )
            ]
        ),
        {"nics": "mac"},
    )

    # NOTE: nics.0 is a vrrp nic with correct mac (check mac == vrid based -> unchanged)
    assert ret["nics"][0]["mac"] == "00:00:5e:00:01:01"

    # NOTE: nics.1 is a vrrp nic with incorrect mac (check mac == vrid based -> changed)
    assert ret["nics"][1]["mac"] == "00:00:5e:00:01:f0"

    # NOTE: nics.2 was not a vrrp nic (check mac was not changed)
    assert ret["nics"][2]["mac"] == "00:22:06:00:00:01"



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import salt.modules.supervisord as supervisord
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SupervisordTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.supervisord
    """

    def setup_loader_modules(self):
        return {supervisord: {}}

    @staticmethod
    def _m_all(stdout=True):
        """
        Return value for cmd.run_all.
        """
        return MagicMock(return_value={"retcode": 0, "stdout": stdout})

    @staticmethod
    def _m_bin():
        """
        Return value for cmd.which_bin.
        """
        return MagicMock(return_value="/tmp/bin_env")

    # 'start' function tests: 1

    def test_start(self):
        """
        Tests if it start the named service.
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.start())

    # 'restart' function tests: 1

    def test_restart(self):
        """
        Tests if it restart the named service.
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.restart())

    # 'stop' function tests: 1

    def test_stop(self):
        """
        Tests if it stop the named service.
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.stop())

    # 'add' function tests: 1

    def test_add(self):
        """
        Tests if it activates any updates in config for process/group.
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.add("salt"))

    # 'remove' function tests: 1

    def test_remove(self):
        """
        Tests if it removes process/group from active config
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.remove("salt"))

    # 'reread' function tests: 1

    def test_reread(self):
        """
        Tests if it reload the daemon's configuration files
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.reread())

    # 'update' function tests: 1

    def test_update(self):
        """
        Tests if it reload config and add/remove as necessary
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.update())

    # 'status' function tests: 1

    def test_status(self):
        """
        Tests if it list programs and its state
        """
        with patch.dict(
            supervisord.__salt__,
            {
                "cmd.run_all": self._m_all("salt running"),
                "cmd.which_bin": self._m_bin(),
            },
        ):
            self.assertDictEqual(
                supervisord.status(), {"salt": {"state": "running", "reason": ""}}
            )

    # 'status_raw' function tests: 1

    def test_status_raw(self):
        """
        Tests if it display the raw output of status
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.status_raw())

    # 'custom' function tests: 1

    def test_custom(self):
        """
        Tests if it run any custom supervisord command
        """
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(supervisord.custom("mstop '*gunicorn*'"))

    # 'options' function tests: 1

    def test_options(self):
        """
        Tests if it read the config file and return the config options
        for a given process
        """

        class MockConfig:
            """
            Mock Config class
            """

            flag = None

            def __init__(self):
                self.name = None

            def sections(self):
                """
                Mock sections methos
                """
                if self.flag == 1:
                    return []
                return ["program:salt"]

            def items(self, name):
                """
                Mock sections methos
                """
                self.name = name
                return [("salt", "True")]

        with patch.object(
            supervisord, "_read_config", MagicMock(return_value=MockConfig())
        ):
            MockConfig.flag = 1
            self.assertRaises(CommandExecutionError, supervisord.options, "salt")

            MockConfig.flag = 0
            self.assertDictEqual(supervisord.options("salt"), {"salt": True})

    # 'status_bool' function tests: 4

    def test_status_bool_name_correct(self):
        """
        Tests if it returns bool results for process status check
        """

        return_str = (
            "salt-minion                     RUNNING   pid 25563, uptime 4 days,"
            " 12:39:42"
        )
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(return_str), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(
                supervisord.status_bool(name="salt-minion", expected_state=None)
            )

    def test_status_bool_name_wrong(self):
        """
        Tests if it returns bool results for process status check
        """

        return_str = (
            "salt-master                     RUNNING   pid 25563, uptime 4 days,"
            " 12:39:42"
        )
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(return_str), "cmd.which_bin": self._m_bin()},
        ):
            self.assertFalse(
                supervisord.status_bool(name="salt-minion", expected_state=None)
            )

    def test_status_bool_expected_state_running_success(self):
        """
        Tests if it returns bool results for process status check
        """

        return_str = (
            "salt-minion                     RUNNING   pid 25563, uptime 4 days,"
            " 12:39:42"
        )
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(return_str), "cmd.which_bin": self._m_bin()},
        ):
            self.assertTrue(
                supervisord.status_bool(name="salt-minion", expected_state="RUNNING")
            )

    def test_status_bool_expected_state_running_fail(self):
        """
        Tests if it returns bool results for process status check
        """

        return_str = (
            "salt-minion                     STOPPED   pid 25563, uptime 4 days,"
            " 12:39:42"
        )
        with patch.dict(
            supervisord.__salt__,
            {"cmd.run_all": self._m_all(return_str), "cmd.which_bin": self._m_bin()},
        ):
            self.assertFalse(
                supervisord.status_bool(name="salt-minion", expected_state="RUNNING")
            )



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.states.virtualenv_mod
"""

import os

import pytest

import salt.states.virtualenv_mod as virtualenv_mod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {virtualenv_mod: {"__env__": "base"}}


def test_managed():
    """
    Test to create a virtualenv and optionally manage it with pip
    """
    ret = {"name": "salt", "changes": {}, "result": False, "comment": ""}
    ret.update({"comment": "Virtualenv was not detected on this system"})
    assert virtualenv_mod.managed("salt") == ret

    mock1 = MagicMock(return_value="True")
    mock = MagicMock(return_value=False)
    mock2 = MagicMock(return_value="1.1")
    with patch.dict(
        virtualenv_mod.__salt__,
        {
            "virtualenv.create": True,
            "cp.is_cached": mock,
            "cp.cache_file": mock,
            "cp.hash_file": mock,
            "pip.freeze": mock1,
            "cmd.run_stderr": mock1,
            "pip.version": mock2,
        },
    ):
        mock = MagicMock(side_effect=[True, True, True, False, True, True])
        with patch.object(os.path, "exists", mock):
            ret.update({"comment": "pip requirements file 'salt://a' not found"})
            assert virtualenv_mod.managed("salt", None, "salt://a") == ret

            with patch.dict(virtualenv_mod.__opts__, {"test": True}):
                ret.update(
                    {
                        "changes": {"cleared_packages": "True", "old": "True"},
                        "comment": "Virtualenv salt is set to be cleared",
                        "result": None,
                    }
                )
                assert virtualenv_mod.managed("salt", clear=1) == ret
                ret.update(
                    {
                        "comment": "Virtualenv salt is already created",
                        "changes": {},
                        "result": True,
                    }
                )
                assert virtualenv_mod.managed("salt") == ret

                ret.update(
                    {
                        "comment": "Virtualenv salt is set to be created",
                        "result": None,
                    }
                )
                assert virtualenv_mod.managed("salt") == ret

            with patch.dict(virtualenv_mod.__opts__, {"test": False}):
                ret.update(
                    {
                        "comment": (
                            "The 'use_wheel' option is"
                            " only supported in pip between 1.4 and 9.0.3."
                            " The version of pip detected was 1.1."
                        ),
                        "result": False,
                    }
                )
                assert virtualenv_mod.managed("salt", use_wheel=1) == ret

                ret.update({"comment": "virtualenv exists", "result": True})
                assert virtualenv_mod.managed("salt") == ret



import pytest

import salt.utils.win_pdh as win_pdh
from tests.support.mock import patch
from tests.support.unit import TestCase

try:
    import pywintypes

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False


@pytest.mark.skipif(not HAS_WIN32, reason="Requires pywin32")
@pytest.mark.skip_unless_on_windows
class WinPdhTestCase(TestCase):
    @pytest.mark.slow_test
    def test_list_objects(self):
        known_objects = ["Cache", "Memory", "Process", "Processor", "System"]
        objects = win_pdh.list_objects()
        for item in known_objects:
            self.assertTrue(item in objects)

    def test_list_counters(self):
        counters = win_pdh.list_counters("Processor")
        known_counters = ["% Processor Time", "% User Time", "% DPC Time"]
        for item in known_counters:
            self.assertTrue(item in counters)

    def test_list_instances(self):
        instances = win_pdh.list_instances("Processor")
        known_instances = ["0", "_Total"]
        for item in known_instances:
            self.assertTrue(item in instances)

    def test_build_counter_list(self):
        counter_list = [
            ("Memory", None, "Available Bytes"),
            ("Paging File", "*", "% Usage"),
            ("Processor", "*", "% Processor Time"),
            ("Server", None, "Work Item Shortages"),
            ("Server Work Queues", "*", "Queue Length"),
            ("System", None, "Context Switches/sec"),
        ]
        resulting_list = win_pdh.build_counter_list(counter_list)
        for counter in resulting_list:
            self.assertTrue(isinstance(counter, win_pdh.Counter))

        resulting_paths = []
        for counter in resulting_list:
            resulting_paths.append(counter.path)

        expected_paths = [
            "\\Memory\\Available Bytes",
            "\\Paging File(*)\\% Usage",
            "\\Processor(*)\\% Processor Time",
            "\\Server\\Work Item Shortages",
            "\\Server Work Queues(*)\\Queue Length",
            "\\System\\Context Switches/sec",
        ]
        self.assertEqual(resulting_paths, expected_paths)

    @pytest.mark.slow_test
    def test_get_all_counters(self):
        results = win_pdh.get_all_counters("Processor")
        known_counters = [
            "\\Processor(*)\\% Processor Time",
            "\\Processor(*)\\% Idle Time",
            "\\Processor(*)\\DPC Rate",
            "\\Processor(*)\\% Privileged Time",
            "\\Processor(*)\\DPCs Queued/sec",
            "\\Processor(*)\\% Interrupt Time",
            "\\Processor(*)\\Interrupts/sec",
        ]
        for item in known_counters:
            self.assertTrue(item in results)

    @pytest.mark.slow_test
    def test_get_counters(self):
        counter_list = [
            ("Memory", None, "Available Bytes"),
            ("Paging File", "*", "% Usage"),
            ("Processor", "*", "% Processor Time"),
            ("Server", None, "Work Item Shortages"),
            ("Server Work Queues", "*", "Queue Length"),
            ("System", None, "Context Switches/sec"),
        ]
        results = win_pdh.get_counters(counter_list)
        expected_counters = [
            "\\Memory\\Available Bytes",
            "\\Paging File(*)\\% Usage",
            "\\Processor(*)\\% Processor Time",
            "\\Server\\Work Item Shortages",
            "\\Server Work Queues(*)\\Queue Length",
            "\\System\\Context Switches/sec",
        ]
        for item in expected_counters:
            self.assertTrue(item in results)

    def test_get_counter(self):
        results = win_pdh.get_counter("Processor", "*", "% Processor Time")
        self.assertTrue("\\Processor(*)\\% Processor Time" in results)

    @patch("win32pdh.CollectQueryData")
    def test_get_counters_no_data_to_return(self, mock_query):
        mock_query.side_effect = pywintypes.error(
            -2147481643, "CollectQueryData", "No data to return."
        )
        counter_list = [
            ("Memory", None, "Available Bytes"),
            ("Paging File", "*", "% Usage"),
            ("Processor", "*", "% Processor Time"),
            ("Server", None, "Work Item Shortages"),
            ("Server Work Queues", "*", "Queue Length"),
            ("System", None, "Context Switches/sec"),
        ]
        results = win_pdh.get_counters(counter_list)
        assert results == {}



"""
Manage macOS local directory passwords and policies

.. versionadded:: 2016.3.0

Note that it is usually better to apply password policies through the creation
of a configuration profile.
"""

# Authentication concepts reference:
# https://developer.apple.com/library/mac/documentation/Networking/Conceptual/Open_Directory/openDirectoryConcepts/openDirectoryConcepts.html#//apple_ref/doc/uid/TP40000917-CH3-CIFCAIBB

import logging
from datetime import datetime

import salt.utils.mac_utils
import salt.utils.platform
from salt.exceptions import CommandExecutionError

try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


log = logging.getLogger(__name__)  # Start logging

__virtualname__ = "shadow"


def __virtual__():
    # Is this macOS?
    if not salt.utils.platform.is_darwin():
        return False, "Not macOS"

    if HAS_PWD:
        return __virtualname__
    else:
        return (False, "The pwd module failed to load.")


def _get_account_policy(name):
    """
    Get the entire accountPolicy and return it as a dictionary. For use by this
    module only

    :param str name: The user name

    :return: a dictionary containing all values for the accountPolicy
    :rtype: dict

    :raises: CommandExecutionError on user not found or any other unknown error
    """

    cmd = f"pwpolicy -u {name} -getpolicy"
    try:
        ret = salt.utils.mac_utils.execute_return_result(cmd)
    except CommandExecutionError as exc:
        if f"Error: user <{name}> not found" in exc.strerror:
            raise CommandExecutionError(f"User not found: {name}")
        raise CommandExecutionError(f"Unknown error: {exc.strerror}")

    try:
        policy_list = ret.split("\n")[1].split(" ")
        policy_dict = {}
        for policy in policy_list:
            if "=" in policy:
                key, value = policy.split("=")
                policy_dict[key] = value
        return policy_dict
    except IndexError:
        return {}


def _set_account_policy(name, policy):
    """
    Set a value in the user accountPolicy. For use by this module only

    :param str name: The user name
    :param str policy: The policy to apply

    :return: True if success, otherwise False
    :rtype: bool

    :raises: CommandExecutionError on user not found or any other unknown error
    """
    cmd = f'pwpolicy -u {name} -setpolicy "{policy}"'

    try:
        return salt.utils.mac_utils.execute_return_success(cmd)
    except CommandExecutionError as exc:
        if f"Error: user <{name}> not found" in exc.strerror:
            raise CommandExecutionError(f"User not found: {name}")
        raise CommandExecutionError(f"Unknown error: {exc.strerror}")


def _get_account_policy_data_value(name, key):
    """
    Return the value for a key in the accountPolicy section of the user's plist
    file. For use by this module only

    :param str name: The username
    :param str key: The accountPolicy key

    :return: The value contained within the key
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error
    """
    cmd = f"dscl . -readpl /Users/{name} accountPolicyData {key}"
    try:
        ret = salt.utils.mac_utils.execute_return_result(cmd)
    except CommandExecutionError as exc:
        if "eDSUnknownNodeName" in exc.strerror:
            raise CommandExecutionError(f"User not found: {name}")
        if "eDSUnknownMatchType" in exc.strerror:
            raise CommandExecutionError(f"Value not found: {key}")
        raise CommandExecutionError(f"Unknown error: {exc.strerror}")

    return ret


def _convert_to_datetime(unix_timestamp):
    """
    Converts a unix timestamp to a human readable date/time

    :param float unix_timestamp: A unix timestamp

    :return: A date/time in the format YYYY-mm-dd HH:MM:SS
    :rtype: str
    """
    try:
        unix_timestamp = float(unix_timestamp)
        return datetime.fromtimestamp(unix_timestamp).strftime("%Y-%m-%d %H:%M:%S")
    except (ValueError, TypeError):
        return "Invalid Timestamp"


def info(name):
    """
    Return information for the specified user

    :param str name: The username

    :return: A dictionary containing the user's shadow information
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.info admin
    """
    try:
        data = pwd.getpwnam(name)
        return {
            "name": data.pw_name,
            "passwd": data.pw_passwd,
            "account_created": get_account_created(name),
            "login_failed_count": get_login_failed_count(name),
            "login_failed_last": get_login_failed_last(name),
            "lstchg": get_last_change(name),
            "max": get_maxdays(name),
            "expire": get_expire(name),
            "change": get_change(name),
            "min": "Unavailable",
            "warn": "Unavailable",
            "inact": "Unavailable",
        }

    except KeyError:
        log.debug("User not found: %s", name)
        return {
            "name": "",
            "passwd": "",
            "account_created": "",
            "login_failed_count": "",
            "login_failed_last": "",
            "lstchg": "",
            "max": "",
            "expire": "",
            "change": "",
            "min": "",
            "warn": "",
            "inact": "",
        }


def get_account_created(name):
    """
    Get the date/time the account was created

    :param str name: The username of the account

    :return: The date/time the account was created (yyyy-mm-dd hh:mm:ss) or 0 if
        the value is not defined
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_account_created admin
    """
    try:
        ret = _get_account_policy_data_value(name, "creationTime")
    except CommandExecutionError as exc:
        if "Value not found" in exc.message:
            return "0"
        else:
            raise

    unix_timestamp = salt.utils.mac_utils.parse_return(ret)
    return _convert_to_datetime(unix_timestamp)


def get_last_change(name):
    """
    Get the date/time the account was changed

    :param str name: The username of the account

    :return: The date/time the account was modified (yyyy-mm-dd hh:mm:ss) or 0
        if the value is not defined
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_last_change admin
    """
    try:
        ret = _get_account_policy_data_value(name, "passwordLastSetTime")
    except CommandExecutionError as exc:
        if "Value not found" in exc.message:
            return "0"
        else:
            raise

    unix_timestamp = salt.utils.mac_utils.parse_return(ret)
    return _convert_to_datetime(unix_timestamp)


def get_login_failed_count(name):
    """
    Get the number of failed login attempts

    :param str name: The username of the account

    :return: The number of failed login attempts. 0 may mean there are no failed
        login attempts or the value is not defined
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_login_failed_count admin
    """
    try:
        ret = _get_account_policy_data_value(name, "failedLoginCount")
    except CommandExecutionError as exc:
        if "Value not found" in exc.message:
            return "0"
        else:
            raise
    return salt.utils.mac_utils.parse_return(ret)


def get_login_failed_last(name):
    """
    Get the date/time of the last failed login attempt

    :param str name: The username of the account

    :return: The date/time of the last failed login attempt on this account
        (yyyy-mm-dd hh:mm:ss) or 0 if the value is not defined
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_login_failed_last admin
    """
    try:
        ret = _get_account_policy_data_value(name, "failedLoginTimestamp")
    except CommandExecutionError as exc:
        if "Value not found" in exc.message:
            return "0"
        else:
            raise

    unix_timestamp = salt.utils.mac_utils.parse_return(ret)
    return _convert_to_datetime(unix_timestamp)


def set_maxdays(name, days):
    """
    Set the maximum age of the password in days

    :param str name: The username of the account

    :param int days: The maximum age of the account in days

    :return: True if successful, False if not
    :rtype: bool

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_maxdays admin 90
    """
    minutes = days * 24 * 60

    _set_account_policy(name, f"maxMinutesUntilChangePassword={minutes}")

    return get_maxdays(name) == days


def get_maxdays(name):
    """
    Get the maximum age of the password

    :param str name: The username of the account

    :return: The maximum age of the password in days
    :rtype: int

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_maxdays admin 90
    """
    policies = _get_account_policy(name)

    if "maxMinutesUntilChangePassword" in policies:
        max_minutes = policies["maxMinutesUntilChangePassword"]
        return int(max_minutes) / 24 / 60

    return 0


def set_mindays(name, days):
    """
    Set the minimum password age in days. Not available in macOS.

    :param str name: The user name

    :param int days: The number of days

    :return: Will always return False until macOS supports this feature.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_mindays admin 90
    """
    return False


def set_inactdays(name, days):
    """
    Set the number if inactive days before the account is locked. Not available
    in macOS

    :param str name: The user name

    :param int days: The number of days

    :return: Will always return False until macOS supports this feature.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_inactdays admin 90
    """
    return False


def set_warndays(name, days):
    """
    Set the number of days before the password expires that the user will start
    to see a warning. Not available in macOS

    :param str name: The user name

    :param int days: The number of days

    :return: Will always return False until macOS supports this feature.
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_warndays admin 90
    """
    return False


def set_change(name, date):
    """
    Sets the date on which the password expires. The user will be required to
    change their password. Format is mm/dd/yyyy

    :param str name: The name of the user account

    :param date date: The date the password will expire. Must be in mm/dd/yyyy
        format.

    :return: True if successful, otherwise False
    :rtype: bool

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_change username 09/21/2016
    """
    _set_account_policy(name, f"usingExpirationDate=1 expirationDateGMT={date}")

    return get_change(name) == date


def get_change(name):
    """
    Gets the date on which the password expires

    :param str name: The name of the user account

    :return: The date the password will expire
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_change username
    """
    policies = _get_account_policy(name)

    if "expirationDateGMT" in policies:
        return policies["expirationDateGMT"]

    return "Value not set"


def set_expire(name, date):
    """
    Sets the date on which the account expires. The user will not be able to
    login after this date. Date format is mm/dd/yyyy

    :param str name: The name of the user account

    :param datetime date: The date the account will expire. Format must be
        mm/dd/yyyy.

    :return: True if successful, False if not
    :rtype: bool

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.set_expire username 07/23/2015
    """
    _set_account_policy(name, f"usingHardExpirationDate=1 hardExpireDateGMT={date}")

    return get_expire(name) == date


def get_expire(name):
    """
    Gets the date on which the account expires

    :param str name: The name of the user account

    :return: The date the account expires
    :rtype: str

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.get_expire username
    """
    policies = _get_account_policy(name)

    if "hardExpireDateGMT" in policies:
        return policies["hardExpireDateGMT"]

    return "Value not set"


def del_password(name):
    """
    Deletes the account password

    :param str name: The user name of the account

    :return: True if successful, otherwise False
    :rtype: bool

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' shadow.del_password username
    """
    # This removes the password
    cmd = f"dscl . -passwd /Users/{name} ''"
    try:
        salt.utils.mac_utils.execute_return_success(cmd)
    except CommandExecutionError as exc:
        if "eDSUnknownNodeName" in exc.strerror:
            raise CommandExecutionError(f"User not found: {name}")
        raise CommandExecutionError(f"Unknown error: {exc.strerror}")

    # This is so it looks right in shadow.info
    cmd = f"dscl . -create /Users/{name} Password '*'"
    salt.utils.mac_utils.execute_return_success(cmd)

    return info(name)["passwd"] == "*"


def set_password(name, password):
    """
    Set the password for a named user (insecure, the password will be in the
    process list while the command is running)

    :param str name: The name of the local user, which is assumed to be in the
        local directory service

    :param str password: The plaintext password to set

    :return: True if successful, otherwise False
    :rtype: bool

    :raises: CommandExecutionError on user not found or any other unknown error

    CLI Example:

    .. code-block:: bash

        salt '*' mac_shadow.set_password macuser macpassword
    """
    cmd = f"dscl . -passwd /Users/{name} '{password}'"
    try:
        salt.utils.mac_utils.execute_return_success(cmd)
    except CommandExecutionError as exc:
        if "eDSUnknownNodeName" in exc.strerror:
            raise CommandExecutionError(f"User not found: {name}")
        raise CommandExecutionError(f"Unknown error: {exc.strerror}")

    return True



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>

    Test cases for salt.modules.lvs
"""

import pytest

import salt.modules.lvs as lvs
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {lvs: {}}


def test_add_service():
    """
    Test for Add a virtual service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.add_service() == "stderr"


def test_edit_service():
    """
    Test for Edit the virtual service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.edit_service() == "stderr"


def test_delete_service():
    """
    Test for Delete the virtual service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.delete_service() == "stderr"


def test_add_server():
    """
    Test for Add a real server to a virtual service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.add_server() == "stderr"


def test_edit_server():
    """
    Test for Edit a real server to a virtual service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.edit_server() == "stderr"


def test_delete_server():
    """
    Test for Delete the realserver from the virtual service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.delete_server() == "stderr"


def test_clear():
    """
    Test for Clear the virtual server table
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.dict(
            lvs.__salt__,
            {
                "cmd.run_all": MagicMock(
                    return_value={"retcode": "ret", "stderr": "stderr"}
                )
            },
        ):
            assert lvs.clear() == "stderr"


def test_get_rules():
    """
    Test for Get the virtual server rules
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.dict(lvs.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            assert lvs.get_rules() == "A"


def test_list_():
    """
    Test for List the virtual server table
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.list_("p", "s") == "stderr"


def test_zero():
    """
    Test for Zero the packet, byte and rate counters in a
     service or all services.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                assert lvs.zero("p", "s") == "stderr"


def test_check_service():
    """
    Test for Check the virtual service exists.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                with patch.object(lvs, "get_rules", return_value="C"):
                    assert lvs.check_service("p", "s") == "Error: service not exists"


def test_check_server():
    """
    Test for Check the real server exists in the specified service.
    """
    with patch.object(lvs, "__detect_os", return_value="C"):
        with patch.object(lvs, "_build_cmd", return_value="B"):
            with patch.dict(
                lvs.__salt__,
                {
                    "cmd.run_all": MagicMock(
                        return_value={"retcode": "ret", "stderr": "stderr"}
                    )
                },
            ):
                with patch.object(lvs, "get_rules", return_value="C"):
                    assert lvs.check_server("p", "s") == "Error: server not exists"
